import click


option = click.option(
    '-P/ ', '--no-prompt/ ', 'configure_prompt',
    default=False,
    show_default=True,
    is_flag=True,
    help='Do not prompt for any missing data. Used when programmatically running `configure [tenant|global]`.'
)



import os
from pathlib import Path
import json


def test_clear_cache(runner, cli):
    result = runner.invoke(cli, 'clear cache'.split(' '))
    local_home = os.getenv('PYBRITIVE_HOME_DIR')
    path = Path(Path(local_home) / '.britive' / 'pybritive.cache')
    with open(str(path), 'r', encoding='utf-8') as f:
        data = json.loads(f.read())
    assert result.exit_code == 0
    assert 'profiles' in data.keys()
    assert len(data['profiles']) == 0


def test_clear_gcloud_key_files(runner, cli):
    local_home = os.getenv('PYBRITIVE_HOME_DIR')
    path = Path(local_home) / '.britive' / 'pybritive-gcloud-key-files'
    file1 = path / 'file1.json'
    file2 = path / 'file2.json'

    path.mkdir(exist_ok=True, parents=True)
    file1.write_text('')
    file2.write_text('')

    assert path.is_dir()
    assert file1.is_file()
    assert file2.is_file()

    result = runner.invoke(cli, 'clear gcloud-auth-key-files'.split(' '))
    assert result.exit_code == 0
    assert not file1.is_file()
    assert not file2.is_file()
    assert not path.is_dir()



def get_args():
    from getopt import getopt  # lazy load
    from sys import argv  # lazy load
    options = getopt(argv[1:], 't:T:p:F:hv', [
        'tenant=',
        'token=',
        'passphrase=',
        'federation-provider=',
        'help',
        'version'
    ])[0]

    args = {
        'tenant': None,
        'token': None,
        'passphrase': None,
        'federation_provider': None
    }

    for opt, arg in options:
        if opt in ('-t', '--tenant'):
            args['tenant'] = arg
        if opt in ('-T', '--token'):
            args['token'] = arg
        if opt in ('-p', '--passphrase'):
            args['passphrase'] = arg
        if opt in ('-F', '--federation-provider'):
            args['federation_provider'] = arg
        if opt in ('-h', '--help'):
            usage()
        if opt in ('-v', '--version'):
            from platform import platform, python_version  # lazy load
            from pkg_resources import get_distribution  # lazy load
            cli_version = get_distribution('pybritive').version
            print(
                f'pybritive: {cli_version} / platform: {platform()} / python: {python_version()}'
            )
            exit()

    return args


def usage():
    from sys import argv  # lazy load
    print(f"Usage : {argv[0]} [-t/--tenant, -T/--token, -t/--passphrase, -F/--federation-provider]")
    exit()


def main():
    args = get_args()

    from .k8s_exec_credential_builder import KubernetesExecCredentialProcessor

    k8s_processor = KubernetesExecCredentialProcessor()

    from .cache import Cache  # lazy load
    creds = Cache(passphrase=args['passphrase']).get_credentials(
        profile_name=k8s_processor.profile,
        mode='kube-exec'
    )
    if creds:
        from datetime import datetime  # lazy load
        expiration = datetime.fromisoformat(creds['expirationTime'].replace('Z', ''))
        now = datetime.utcnow()
        if now > expiration:  # creds have expired so set to none so new one get checked out
            creds = None
        else:
            print(k8s_processor.construct_exec_credential(creds))
            exit()

    if not creds:
        from ..britive_cli import BritiveCli  # lazy load for performance purposes

        b = BritiveCli(
            tenant_name=args['tenant'],
            token=args['token'],
            passphrase=args['passphrase'],
            federation_provider=args['federation_provider'],
            silent=True,
            from_helper_console_script=True
        )
        b.config.get_tenant()  # have to load the config here as that work is generally done elsewhere
        b.checkout(
            alias=None,
            blocktime=None,
            console=False,
            justification=None,
            mode='kube-exec',
            maxpolltime=None,
            profile=k8s_processor.profile,
            passphrase=args['passphrase'],
            force_renew=None,
            aws_credentials_file=None,
            gcloud_key_file=None,
            verbose=None,
            extend=False,
            otp=None,
        )
        exit()


if __name__ == '__main__':
    main()



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options


@click.group()
def aws():
    """Groups together AWS specific sub-commands. Naming conventions will change to support standard AWS names."""
    pass


@aws.command()
@build_britive
@britive_options(names='aws_profile,aws_console_duration,browser')
def console(ctx, profile, duration, browser):
    """Signs an AWS console federation URL with AWS access keys and opens the AWS console in a browser."""

    ctx.obj.britive.aws_console(
        profile=profile,
        duration=duration,
        browser=browser
    )



import click

# eval example: eval $(pybritive checkout test -m env)

mode_choices = click.Choice(
    [
        'text',                     # text based output
        'json',                     # json based output
        'env',                      # environment variable output (suitable for eval)
        'integrate',                # stick the credentials into the local CSP credentials file (AWS)
        'env-nix',                  # environment variable output specifying "export"
        'env-wincmd',               # environment variable output specifying "set"
        'env-winps',                # environment variable output specifying "$env:"
        'awscredentialprocess',     # aws credential process output with additional caching to make the credential process more performant
        'azlogin',                  # azure az login command with all fields populated (suitable for eval)
        'azps',                     # azure powershell script
        'browser',                  # checkout console access (without having to specify --console/-c) and open the default browser to the URL provided in the checkout
        'gcloudauth',               # gcloud auth activate-service-account with all fields populated (suitable for eval)
        'console',                  # checkout console access (without having to specify --console/-c) and print the URL
        'browser-mozilla',
        'browser-firefox',
        'browser-windows-default',
        'browser-macosx',
        'browser-safari',
        'browser-chrome',
        'browser-chromium',
        'kube-exec',                # bake into kubeconfig with oidc exec output and additional caching to make kubectl more performant
        'gcloudauthexec',           # will effectively execute results of gcloudauth in a sub-shell
        'os-oclogin',               # will attempt an oidc authorization code grant flow for generating the `oc login ...` command for OpenShift
        'os-ocloginexec',           # will attempt an oidc authorization code grant flow for generating the `oc login ...` command for OpenShift and exec the result in a subshell
    ],
    case_sensitive=False
)



import click

ssh_key_source_choices = click.Choice(
    [
        'ssh-agent',    # use the ssh-agent and ssh-add the key so the IdentityFile parameter can be omitted
        'static',       # use a well known key name format so IdentityFile parameter can be used
    ],
    case_sensitive=False
)



# register the completion scripts since they are not required to be registered anywhere else

import pkg_resources

# click < 8.0.0 does shell completion different...
# not all the classes/decorators are available, so we cannot
# create custom shell completions like we can with click > 8.0.0
click_major_version = int(pkg_resources.get_distribution('click').version.split('.')[0])
if click_major_version >= 8:
    from . import powershell_completion
    from . import bash_gte_42



import click

output_format_choices = click.Choice(
    [
        'json',
        'yaml',
        'csv',
        'table',
        'list',
        'table-plain',
        'table-simple',
        'table-github',
        'table-grid',
        'table-simple_grid',
        'table-rounded_grid',
        'table-double_grid',
        'table-fancy_grid',
        'table-outline',
        'table-simple_outline',
        'table-rounded_outline',
        'table-double_outline',
        'table-fancy_outline',
        'table-pipe',
        'table-orgtbl',
        'table-jira',
        'table-presto',
        'table-pretty',
        'table-psql',
        'table-rst',
        'table-mediawiki',
        'table-moinmoin',
        'table-youtrack',
        'table-html',
        'table-unsafehtml',
        'table-latex',
        'table-latex_raw',
        'table-latex_booktabs',
        'table-latex_longtable',
        'table-textile',
        'table-tsv'
    ],
    case_sensitive=False
)



from pathlib import Path
import os
import warnings


def test_login_interactive(runner, cli):
    result = runner.invoke(cli, ['login'])
    api_login_warning = 'Interactive login unavailable when an API token is provided.'
    if api_login_warning in result.output:
        warnings.warn(api_login_warning.upper())
        return
    assert result.exit_code == 0
    local_home = os.getenv('PYBRITIVE_HOME_DIR')
    path = Path(Path(local_home) / '.britive' / 'pybritive.credentials.encrypted')
    with open(str(path), 'r', encoding='utf-8') as f:
        data = f.read()
    assert 'accessToken=' in data
    assert os.getenv('PYBRITIVE_TEST_TENANT') in data



import configparser
import json
import os
import subprocess
from pathlib import Path
import platform
import uuid
import webbrowser
import click


# trailing spaces matter as some options do not have the trailing space
env_options = {
    'nix': 'export ',
    'winps': '$Env:',
    'wincmd': 'set '
}


def safe_list_get(l, idx, default):
    try:
        return l[idx]
    except IndexError:
        return default


class CloudCredentialPrinter:
    def __init__(self, app_type, console, mode, profile, silent, credentials, cli):
        self.cli = cli
        self.silent = silent
        self.app_type = app_type
        self.profile = profile
        self.console = console
        mode = mode or 'json'  # set a default if nothing is provided via flag --mode/-m
        self.full_mode = mode
        helper = mode.split('-')
        self.mode = helper[0]
        self.mode_modifier = safe_list_get(mode.split('-', maxsplit=1), 1, None)
        self.credentials = credentials
        if self.mode == 'env':
            if self.mode_modifier:
                self.env_command = env_options[self.mode_modifier]
            else:
                self.on_windows = platform.system().lower() == 'windows'
                self.env_command = env_options['wincmd'] if self.on_windows else env_options['nix']
        self.modes_to_skip_console_printing = ['os-oclogin', 'os-ocloginexec']

    def print(self):
        if self.console and self.full_mode not in self.modes_to_skip_console_printing:
            self.print_console()
            return
        if self.mode == 'text':
            self.print_text()
        if self.mode == 'json':
            self.print_json()
        if self.mode == 'env':
            self.print_env()
        if self.mode == 'integrate':
            self.print_integrate()
        if self.mode == 'azlogin':
            self.print_azlogin()
        if self.mode == 'awscredentialprocess':
            self.print_awscredentialprocess()
        if self.mode == 'azps':
            self.print_azps()
        if self.mode == 'gcloudauth':
            self.print_gcloudauth()
        if self.mode == 'kube':
            self.print_kube()
        if self.mode == 'os':
            self.print_os()
        if self.mode == 'gcloudauthexec':
            self.print_gcloudauthexec()

    def print_console(self):
        url = self.credentials.get('url', self.credentials)
        if self.mode == 'browser':
            browser = self.mode_modifier or os.getenv('PYBRITIVE_BROWSER')
            webbrowser.get(using=browser).open(url)
        else:
            self.cli.print(url, ignore_silent=True)

    def print_text(self):
        self._not_implemented()

    def print_json(self):
        self._not_implemented()

    def print_env(self):
        self._not_implemented()

    def print_integrate(self):
        self._not_implemented()

    def print_awscredentialprocess(self):
        self._not_implemented()

    def print_azlogin(self):
        self._not_implemented()

    def print_azps(self):
        self._not_implemented()

    def print_gcloudauth(self):
        self._not_implemented()

    def print_kube(self):
        self._not_implemented()

    def print_gcloudauthexec(self):
        self._not_implemented()

    def print_os(self):
        self._not_implemented()

    def _not_implemented(self):
        raise click.ClickException(f'Application type {self.app_type} does not support the specified mode.')


class GenericCloudCredentialPrinter(CloudCredentialPrinter):
    def __init__(self, console, mode, profile, silent, credentials, cli):
        super().__init__('Generic', console, mode, profile, silent, credentials, cli)

    def print_json(self):
        try:
            self.cli.print(json.dumps(self.credentials, indent=2), ignore_silent=True)
        except json.JSONDecodeError:
            self.cli.print(self.credentials, ignore_silent=True)


class ResourcesCredentialPrinter(CloudCredentialPrinter):
    def __init__(self, profile, silent, credentials, cli):
        super().__init__('Resources', False, 'text', profile, silent, credentials, cli)

    def print_text(self):
        self.cli.print(self.credentials, ignore_silent=True)


class AwsCloudCredentialPrinter(CloudCredentialPrinter):
    def __init__(self, console, mode, profile, silent, credentials, cli, aws_credentials_file):
        super().__init__('AWS', console, mode, profile, silent, credentials, cli)
        self.aws_credentials_file = aws_credentials_file

    def print_text(self):
        self.cli.print('AWS_ACCESS_KEY_ID', ignore_silent=True)
        self.cli.print(self.credentials['accessKeyID'], ignore_silent=True)
        self.cli.print('', ignore_silent=True)

        self.cli.print('AWS_SECRET_ACCESS_KEY', ignore_silent=True)
        self.cli.print(self.credentials['secretAccessKey'], ignore_silent=True)
        self.cli.print('', ignore_silent=True)

        self.cli.print('AWS_SESSION_TOKEN', ignore_silent=True)
        self.cli.print(self.credentials['sessionToken'], ignore_silent=True)
        self.cli.print('', ignore_silent=True)

        self.cli.print('AWS_EXPIRATION', ignore_silent=True)
        self.cli.print(self.credentials['expirationTime'], ignore_silent=True)
        self.cli.print('', ignore_silent=True)

    def print_json(self, version=None):
        creds = {
            'AccessKeyId': self.credentials['accessKeyID'],
            'SecretAccessKey': self.credentials['secretAccessKey'],
            'SessionToken': self.credentials['sessionToken'],
            'Expiration': self.credentials['expirationTime']
        }
        if version:
            creds['Version'] = version
        self.cli.print(json.dumps(creds, indent=2), ignore_silent=True)

    def print_env(self):
        self.cli.print(f'{self.env_command}AWS_ACCESS_KEY_ID="{self.credentials["accessKeyID"]}"', ignore_silent=True)
        self.cli.print(f'{self.env_command}AWS_SECRET_ACCESS_KEY="{self.credentials["secretAccessKey"]}"',
                       ignore_silent=True)
        self.cli.print(f'{self.env_command}AWS_SESSION_TOKEN="{self.credentials["sessionToken"]}"', ignore_silent=True)
        self.cli.print(f'{self.env_command}AWS_EXPIRATION="{self.credentials["expirationTime"]}"', ignore_silent=True)

    def print_integrate(self):
        # get path to aws credentials file
        env_path = self.aws_credentials_file
        if not env_path:
            path = Path.home() / '.aws' / 'credentials'  # handle os specific separators properly
        else:
            path = Path(env_path)

        # if credentials file does not yet exist, create it as an empty file
        if not path.is_file():
            path.parent.mkdir(exist_ok=True, parents=True)
            path.write_text('', encoding='utf-8')

        # open the file with configparser
        config = configparser.ConfigParser()
        config.read(str(path))

        # add the new profile/section
        config[self.profile] = {
            'aws_access_key_id': self.credentials["accessKeyID"],
            'aws_secret_access_key': self.credentials["secretAccessKey"],
            'aws_session_token': self.credentials["sessionToken"],
            'aws_expiration': self.credentials["expirationTime"]
        }

        # write the new credentials file
        with open(str(path), 'w', encoding='utf-8') as f:
            config.write(f, space_around_delimiters=False)

    def print_awscredentialprocess(self):
        self.print_json(version=1)


class AzureCloudCredentialPrinter(CloudCredentialPrinter):
    def __init__(self, console, mode, profile, silent, credentials, cli):
        key = list(credentials.keys())[0]
        if key != 'url':  # console url is handled differently than programmatic keys so account for it here
            credentials = json.loads(credentials[key])
        super().__init__('Azure', console, mode, profile, silent, credentials, cli)

    def print_text(self):
        self.cli.print('TENANT', ignore_silent=True)
        self.cli.print(self.credentials['tenantId'], ignore_silent=True)
        self.cli.print('', ignore_silent=True)

        self.cli.print('USERNAME/APP ID/CLIENT ID', ignore_silent=True)
        self.cli.print(self.credentials['appId'], ignore_silent=True)
        self.cli.print('', ignore_silent=True)

        self.cli.print('PASSWORD/SECRET TEXT/CLIENT SECRET', ignore_silent=True)
        self.cli.print(self.credentials['secretText'], ignore_silent=True)
        self.cli.print('', ignore_silent=True)

    def print_json(self):
        creds = {
            'TenantId': self.credentials['tenantId'],
            'ClientId': self.credentials['appId'],
            'ClientSecret': self.credentials['secretText']
        }
        self.cli.print(json.dumps(creds, indent=2), ignore_silent=True)

    def print_env(self):
        self.cli.print(f'{self.env_command}AZURE_CLIENT_ID="{self.credentials["appId"]}"', ignore_silent=True)
        self.cli.print(f'{self.env_command}AZURE_CLIENT_SECRET="{self.credentials["secretText"]}"', ignore_silent=True)
        self.cli.print(f'{self.env_command}AZURE_TENANT_ID="{self.credentials["tenantId"]}"', ignore_silent=True)

    def print_azlogin(self):
        self.cli.print(self.credentials['cliLogin'], ignore_silent=True)

    def print_azps(self):
        self.cli.print(self.credentials['powershellScript'].replace('\n ', '\n'), ignore_silent=True)


class GcpCloudCredentialPrinter(CloudCredentialPrinter):
    def __init__(self, console, mode, profile, silent, credentials, cli, gcloud_key_file):
        key = list(credentials.keys())[0]
        credentials = json.loads(credentials[key]) if key != 'url' else credentials
        super().__init__('GCP', console, mode, profile, silent, credentials, cli)
        self.gcloud_key_file = gcloud_key_file

    def print_json(self):
        self.cli.print(json.dumps(self.credentials, indent=2), ignore_silent=True)
        self.cli.print('', ignore_silent=True)
        self.cli.print(
            f"Run command: gcloud auth activate-service-account {self.credentials['client_email']} "
            "--key-file <path-where-above-json-is-stored>", ignore_silent=True
        )

    def print_gcloudauth(self):
        # get path to gcloud key file
        if not self.gcloud_key_file:  # if --gcloud-key-file not provided
            path = Path(self.cli.config.gcloud_key_file_path) / f'{uuid.uuid4().hex}.json'
        else:  # we need to parse out/sanitize what was provided
            path = Path(self.gcloud_key_file).expanduser().absolute()

        # key file does not yet exist so write to it
        path.parent.mkdir(exist_ok=True, parents=True)
        path.write_text(json.dumps(self.credentials, indent=2), encoding='utf-8')

        self.cli.print(
            f"gcloud auth activate-service-account {self.credentials['client_email']} --key-file {str(path)}",
            ignore_silent=True
        )

    def print_gcloudauthexec(self):
        key_file = self.cli.build_gcloud_key_file_for_gcloudauthexec(profile=self.profile)
        path = Path(self.cli.config.gcloud_key_file_path) / key_file

        # key file does not yet exist so write to it
        path.parent.mkdir(exist_ok=True, parents=True)
        path.write_text(json.dumps(self.credentials, indent=2), encoding='utf-8')

        try:
            commands = [
                'gcloud',
                'auth',
                'activate-service-account',
                self.credentials['client_email'],
                '--key-file',
                str(path),
                '--verbosity=error'
            ]

            subprocess.run(commands, check=True)
        except Exception as e:
            self.cli.print(f'error running `gcloud auth activate-service-account ...`: {str(e)}')


class KubernetesCredentialPrinter(CloudCredentialPrinter):
    def __init__(self, console, mode, profile, silent, credentials, cli, k8s_processor):
        self.k8s_processor = k8s_processor
        super().__init__('Kubernetes', console, mode, profile, silent, credentials, cli)

    def print_json(self):
        try:
            self.cli.print(json.dumps(self.credentials, indent=2), ignore_silent=True)
        except json.JSONDecodeError:
            self.cli.print(self.credentials, ignore_silent=True)

    def print_kube(self):
        if self.mode_modifier == 'exec':
            self.cli.print(self.k8s_processor.construct_exec_credential(self.credentials), ignore_silent=True)
        else:
            raise ValueError(f'--mode modifier {self.mode_modifier} for mode {self.mode} not supported')


class OpenShiftCredentialPrinter(CloudCredentialPrinter):
    def __init__(self, console, mode, profile, silent, credentials, cli):
        super().__init__('OpenShift', console, mode, profile, silent, credentials, cli)

    def print_json(self):
        try:
            self.cli.print(json.dumps(self.credentials, indent=2), ignore_silent=True)
        except json.JSONDecodeError:
            self.cli.print(self.credentials, ignore_silent=True)

    def _perform_oidc_auth_code_grant_flow(self):
        try:
            import requests
            from requests.adapters import HTTPAdapter
            from urllib.parse import urlparse

            try:
                from bs4 import BeautifulSoup
            except ImportError as e:
                raise click.ClickException(self.cli.build_import_exception_message('openshift')) from e

            class BritiveCustomHeaderAdapter(HTTPAdapter):
                def __init__(self, headers=None, *args, **kwargs):
                    self.headers = headers or {}
                    super().__init__(*args, **kwargs)

                def add_headers(self, request, **kwargs):
                    for h, v in self.headers.items():
                        request.headers.setdefault(h, v)

            sdk_headers = self.cli.b.session.headers
            tenant_fqdn = self.cli.b.tenant
            temp_url = self.credentials['url']
            temp_url = temp_url.replace('console-openshift-console', 'oauth-openshift')
            parsed_url = urlparse(temp_url)
            base_url = parsed_url.scheme + "://" + parsed_url.netloc
            idp = self.credentials['idpName']

            session = requests.session()

            # set a reasonable user agent, so we can identify traffic more easily
            session.headers.update({
                'User-Agent': sdk_headers['User-Agent']
            })

            # create a custom adapter for the tenants fqdn and mount it
            # this will include the authorization header, so we can "auto"
            # authenticate to the tenant
            adapter = BritiveCustomHeaderAdapter(headers={'Authorization': sdk_headers['Authorization']})
            session.mount(f'https://{tenant_fqdn}/', adapter)

            token_request_url = f'{base_url}/oauth/token/request'
            idp_selection_url = session.get(token_request_url, allow_redirects=False).headers['Location']
            idp_selected_url = f'{idp_selection_url}&idp={idp}'
            response = session.get(idp_selected_url, allow_redirects=True)
            # we now have a html file that we need to extract some things from
            soup = BeautifulSoup(response.content, 'html.parser')
            input_values = {}
            form = soup.find('form')
            inputs = form.find_all('input')
            for inp in inputs:
                name = inp.get('name')
                value = inp.get('value')
                if name:
                    input_values[name] = value

            form_action = form['action']
            if form_action.startswith('/'):
                form_action = form_action[1:]
            response = session.post(f'{base_url}/{form_action}', data=input_values)

            # and we get more html content we need to parse now
            soup = BeautifulSoup(response.content, 'html.parser')

            command = None
            for line in soup.get_text().split('\n'):
                if line.startswith('oc login'):
                    command = line
                    break

            if command:
                return command
            else:
                raise Exception(f'error: no `oc login` command found')
        except Exception as e:
            self.cli.print(f'error when attempting to perform oidc auth code grant flow: {str(e)}', ignore_silent=True)

    def print_os(self):
        if self.mode_modifier == 'oclogin':
            command = self._perform_oidc_auth_code_grant_flow()
            self.cli.print(command, ignore_silent=True)
        elif self.mode_modifier == 'ocloginexec':
            command = self._perform_oidc_auth_code_grant_flow()
            try:
                subprocess.run(command.split(' '), check=True)
            except Exception as e:
                self.cli.print(f'error running `gcloud auth activate-service-account ...`: {str(e)}')
        else:
            raise ValueError(f'--mode modifier {self.mode_modifier} for mode {self.mode} not supported')



import click


option = click.option(
    '--aws-credentials-file',
    default=None,
    help='AWS Programmatic Only - When mode is `integrate` specify a non-default location for the AWS '
         'credentials file.',
    envvar='AWS_SHARED_CREDENTIALS_FILE',
    show_envvar=True,
    show_default=True
)



import click
from ..choices.ssh_push_public_key import ssh_push_public_key_choices


def validate(ctx, param, value):
    cloud = str(ctx.parent.command.name)
    if value == 'default' and cloud == 'aws':
        return 'ec2-instance-connect'
    if value == 'default' and cloud == 'gcp':
        return 'os-login'

    if cloud == 'aws' and value not in ['ec2-instance-connect', None]:
        raise click.BadParameter(f'value {value} is not allowed for aws')

    if cloud == 'gcp' and value not in ['os-login', 'instance-metadata', None]:
        raise click.BadParameter(f'value {value} is not allowed for gcp')

    return value


option = click.option(
    '--push-public-key', '-P',
    type=ssh_push_public_key_choices,
    is_flag=False,
    flag_value='default',
    required=False,
    help='Whether to push an ephemeral public key to the remote server.',
    callback=validate
)



import os
from pathlib import Path
import json


def test_cache_profiles(runner, cli):
    result = runner.invoke(cli, 'cache profiles'.split(' '))
    local_home = os.getenv('PYBRITIVE_HOME_DIR')
    path = Path(Path(local_home) / '.britive' / 'pybritive.cache')
    with open(str(path), 'r', encoding='utf-8') as f:
        data = json.loads(f.read())
    assert result.exit_code == 0
    assert 'profiles' in data.keys()
    assert len(data['profiles']) > 0
    assert len(data['profiles'][0].split('/')) in [2, 3]



import click


option = click.option(
    '--profile', '-p',
    default=None,
    help='AWS CLI profile from which credentials will be sourced. Defaults to None which indicates that the default '
         'credential provider chain should be used.',
    show_envvar=True,
    show_default=True
)



import click
from ..helpers.build_britive import build_britive
from ..helpers.profile_argument_decorator import click_smart_profile_argument


@click.group()
def clear():
    """Clear various local settings and configurations."""
    pass


@clear.command()
@build_britive
def cache(ctx):
    """Clears the local cache."""
    ctx.obj.britive.cache_clear()


@clear.command(name='kubeconfig')
@build_britive
def clear_kubeconfig(ctx):
    """Clears the local .britive/kube/config file."""
    ctx.obj.britive.clear_kubeconfig()


@clear.command(name='gcloud-auth-key-files')
@build_britive
def gcloud_auth_key_files(ctx):
    """Clears the local gcloud auth key files."""
    ctx.obj.britive.clear_gcloud_auth_key_files()


@clear.command(name='cached-aws-credentials')
@build_britive
@click_smart_profile_argument
def cached_aws_credentials(ctx, profile):
    """Clears cached AWS credentials used as part of the AWS CLI credential process."""
    ctx.obj.britive.clear_cached_aws_credentials(profile=profile)



import os
import click

# eval example: eval $(pybritive checkout test -m env)

browser_choices = click.Choice(
    [
        c
        for c in [
            'mozilla',
            'firefox',
            'windows-default',
            'macosx',
            'safari',
            'chrome',
            'chromium',
            os.getenv('PYBRITIVE_BROWSER'),
        ]
        if c
    ],
    case_sensitive=False,
)



import click


option = click.option(
    '--justification', '-j',
    default=None,
    show_default=True,
    help='Justification for the checkout approval process, if the profile checkout requires approval.',
)



import click


option = click.option(
    '--gcloud-key-file',
    default=None,
    help='GCP Programmatic Only - When mode is `gcloudauth` specify a non-default location for storing the key file. Ignored for mode `gcloudauthexec`.',
    show_default=True
)



import click
from click import Command
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options
from ..completers.api_command import command_api_patch_shell_complete
from ..helpers.api_method_argument_decorator import click_smart_api_method_argument


# this holds all the click version logic to gracefully degrade functionality
# depending on the click version
command_api_patch_shell_complete(Command)


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True
    )
)
@build_britive
@britive_options(names='query,output_format,tenant,token,silent,passphrase,federation_provider')
@click_smart_api_method_argument  # need to gracefully handle older version of click
def api(ctx, query, output_format, tenant, token, silent, passphrase, federation_provider, method):
    """Exposes the Britive Python SDK methods to the CLI.

    Documentation on each SDK method can be found inside the Python SDK itself and on Github
    (https://github.com/britive/python-sdk). The Python package `britive` is a dependency of the CLI
    already so the SDK is available without installing any extra packages.

    It is left up to the caller to provide the proper `method` and `parameters` based on the documentation
    of the API call being performed.

    The authenticated identity must have the appropriate permissions to perform the actions being requested.
    General end users of Britive will not have these permissions. This call (and the larger SDK) is generally
    meant for administrative functionality.

    Example of use:

    * generic: pybritive api method --parameter1 value1 --parameter2 value2 [--parameterX valueX]

    * pybritive api users.list

    * pybritive api tags.create --name testtag --description "test tag"

    * pybritive api users.list --query '[].email'

    * pybritive api profiles.create --application-id <id> --name testprofile

    """
    parameters = {ctx.args[i][2:]: ctx.args[i + 1] for i in range(0, len(ctx.args), 2)}
    ctx.obj.britive.api(
        method=method,
        parameters=parameters,
        query=query
    )



import configparser
import json
import os
from pathlib import Path
import shutil
import toml
from britive.britive import Britive
import click
from ..choices.backend import backend_choices
from ..choices.mode import mode_choices
from ..choices.output_format import output_format_choices
from ..helpers.split import profile_split


def extract_tenant(tenant_key):
    return tenant_key.replace('tenant-', '')


def lowercase(obj):
    """Make dictionary lowercase"""
    if isinstance(obj, dict):
        return {k.lower(): lowercase(v) for k, v in obj.items()}
    if isinstance(obj, (list, set, tuple)):
        t = type(obj)
        return t(lowercase(o) for o in obj)
    if isinstance(obj, str):
        return obj.lower()
    return obj


def coalesce(*arg):
    for el in arg:
        if el is not None:
            return el
    return None


non_tenant_sections = ['global', 'profile-aliases', 'aws', 'gcp']

global_fields = [
    'default_tenant',
    'output_format',
    'credential_backend',
    'auto-refresh-profile-cache',
    'auto-refresh-kube-config',
    'ca_bundle',
]

tenant_fields = ['name', 'output_format', 'sso_idp']

aws_fields = ['default_checkout_mode']

gcp_fields = ['gcloud_default_account']


class ConfigManager:
    def __init__(self, cli: object, tenant_name: str = None):
        self.tenant_name = tenant_name
        self.home = os.getenv('PYBRITIVE_HOME_DIR', str(Path.home()))
        self.base_path = str(Path(self.home) / '.britive')
        self.path = str(Path(self.base_path) / 'pybritive.config')
        self.config = None
        self.alias = None
        self.default_tenant = None
        self.tenants = None
        self.tenants_by_name = None
        self.aliases_and_names = None
        self.profile_aliases = {}
        self.cli: object = cli
        self.loaded = False
        self.validation_error_messages = []
        self.gcloud_key_file_path: str = str(Path(self.path).parent / 'pybritive-gcloud-key-files')
        self.global_ca_bundle = None

    def clear_gcloud_auth_key_files(self, profile=None):
        path = Path(self.gcloud_key_file_path)
        if profile:  # if we are given a specific profile we should clear just that key file
            key_file = self.cli.build_gcloud_key_file_for_gcloudauthexec(profile=profile)
            path = path / key_file

            # path.unlink(missing_ok=True)
            # removed for now, for 3.7 compatability
            try:
                path.unlink()
            except FileNotFoundError:
                pass
        else:  # otherwise we can remove all items in the directory and the directory itself
            shutil.rmtree(str(path), ignore_errors=True)

    def get_output_format(self, output_format: str = None):
        return coalesce(
            output_format,
            self.get_tenant().get('output_format'),
            self.config.get('global', {}).get('output_format'),
            'json',  # set to json if no output format is provided
        )

    def load(self, force=False):
        if self.loaded and not force:
            return
        path = Path(self.path)

        if not path.is_file():  # config file does not yet exist, create it as an empty file
            path.parent.mkdir(exist_ok=True, parents=True)
            path.write_text('', encoding='utf-8')

        config = configparser.ConfigParser()
        config.optionxform = str  # maintain key case
        config.read(str(path))
        config = json.loads(json.dumps(config._sections))  # TODO this is messy but works for now

        self.config = lowercase(config)
        self.alias = None  # will be set in self.get_tenant()
        self.default_tenant = self.config.get('global', {}).get('default_tenant')
        self.tenants = {}
        self.tenants_by_name = {}
        for key in self.config:
            if key.startswith('tenant-'):
                alias = extract_tenant(key)
                item = {**self.config[key], **{'alias': alias}}
                self.tenants[alias] = item

                name = self.config[key].get('name', alias)
                if name != alias:
                    self.tenants_by_name[name] = item
        self.aliases_and_names = {**self.tenants, **self.tenants_by_name}
        self.profile_aliases = self.config.get('profile-aliases', {})
        self.global_ca_bundle = self.config.get('ca_bundle', {})
        self.loaded = True

    def get_tenant(self):
        # load up the config - doing it here instead of __init__ for the configure commands since config won't
        # yet exist and we don't want to error
        self.load()  # will set self.config and other variables

        # normalize the input
        name = self.tenant_name.lower() if self.tenant_name else None

        # do some error checking to ensure we can actually grab a tenant
        if len(self.tenants) == 0 and not name:
            raise click.ClickException(f'No tenants found in {self.path}. Cannot continue.')

        # attempt to determine the name of the tenant based on what the user passed in (or didn't pass in)
        provided_tenant_name = name if name else self.default_tenant

        if not provided_tenant_name:  # name not provided and no default has been set
            if len(self.tenants) != 1:
                raise click.ClickException(
                    'Tenant not provided, no default tenant set, and more than one tenant exists.'
                )
            # nothing given but only 1 tenant so assume that is what should be used
            provided_tenant_name = list(self.tenants)[0]

        # if we get here then we now have a tenant name we can check to ensure exists
        if provided_tenant_name not in self.aliases_and_names and not name:
            raise click.ClickException(f'Tenant name "{provided_tenant_name}" not found in {self.path}')

        self.alias = provided_tenant_name or name
        # return details about the requested tenant
        return self.aliases_and_names.get(provided_tenant_name, {'name': name, 'alias': name})

    def save(self):
        self.validate()  # ensure we are actually writing a valid config
        config = configparser.ConfigParser()
        config.optionxform = str  # maintain key case
        config.read_dict(self.config)

        # write the new credentials file
        with open(str(self.path), 'w', encoding='utf-8') as f:
            config.write(f, space_around_delimiters=False)

    def save_tenant(self, tenant: str, alias: str = None, output_format: str = None):
        self.load()
        if not alias:
            alias = tenant
        if f'tenant-{alias}' not in self.config:
            self.config[f'tenant-{alias}'] = {}
        self.config[f'tenant-{alias}']['name'] = tenant
        if output_format:
            self.config[f'tenant-{alias}']['output_format'] = output_format
        self.save()

    def save_global(self, default_tenant_name: str = None, output_format: str = None, backend: str = None):
        self.load()
        if not default_tenant_name and not output_format and not backend:
            return
        if 'global' not in self.config:
            self.config['global'] = {}
        if default_tenant_name:
            self.config['global']['default_tenant'] = default_tenant_name
        if output_format:
            self.config['global']['output_format'] = output_format
        if backend:
            self.config['global']['credential_backend'] = backend
        self.save()

    def get_profile_aliases(self, reverse_keys: bool = False):
        self.load()
        aliases = self.config.get('profile-aliases', {})
        if reverse_keys:
            return {v: k for k, v in aliases.items()}
        else:
            return aliases

    def save_profile_alias(self, alias, profile):
        self.profile_aliases[alias] = profile
        self.config['profile-aliases'] = self.profile_aliases
        self.save()

    # returns a dict of profile aliases that need to be created after listing profiles
    def import_global_npm_config(self):
        self.load()
        path = str(Path(self.home) / '.britive' / 'config')  # handle os specific separators properly
        with open(path, 'r', encoding='utf-8') as f:
            npm_config = toml.load(f)
        tenant = npm_config.get('tenantURL', '').replace('https://', '').replace('.britive-app.com', '').lower()
        output_format = npm_config.get('output_format', '').lower()
        aws_section = npm_config.get('AWS', None)

        # reset the config as we are building a new one
        self.config = {'global': {}}
        if tenant != '':
            self.cli.print(f'Found tenant {tenant}.')
            self.config['global']['default_tenant'] = tenant
            self.config[f'tenant-{tenant}'] = {'name': tenant}
        if output_format != '':
            self.cli.print(f'Found default output format {output_format}.')
            self.config['global']['output_format'] = output_format

        if aws_section:
            checkout_mode = aws_section.get('checkoutMode')
            if checkout_mode:
                checkout_mode = checkout_mode.lower().replace('display', '')
                self.cli.print(f'Found aws default checkout mode of {checkout_mode}.')
                self.config['aws'] = {'default_checkout_mode': checkout_mode}

        self.save()
        self.load(force=True)

        return npm_config.get('envProfileMap', {})

    def backend(self):
        self.load()
        return self.config.get('global', {}).get('credential_backend', 'encrypted-file')

    def aws_default_checkout_mode(self):
        self.load()
        return self.config.get('aws', {}).get('default_checkout_mode', None)

    def gcloud_default_account(self):
        self.load()
        return self.config.get('gcp', {}).get('gcloud_default_account', None)

    def update(self, section, field, value):
        self.load()
        if section not in self.config:
            self.config[section] = {}
        if field not in self.config[section]:
            self.config[section][field] = ''
        self.config[section][field] = value
        self.save()

    def validate(self):
        self.validation_error_messages = []

        for section, fields in self.config.items():
            if section not in non_tenant_sections and not section.startswith('tenant-'):
                self.validation_error_messages.append(f'Invalid section {section} provided.')
            if section == 'global':
                self.validate_global(section, fields)
            if section == 'profile-aliases':
                self.validate_profile_aliases(section, fields)
            if section == 'aws':
                self.validate_aws(section, fields)
            if section == 'gcp':
                self.validate_gcp(section, fields)
            if section.startswith('tenant-'):
                self.validate_tenant(section, fields)

        if len(self.validation_error_messages) > 0:
            message = 'Cannot save config file due to invalid data provided. Details below.'
            errors = [message] + [f'* {m}' for m in self.validation_error_messages]
            raise click.ClickException('\n'.join(errors))

    def validate_global(self, section, fields):
        for field, value in fields.items():
            if field not in global_fields:
                self.validation_error_messages.append(f'Invalid {section} field {field} provided.')
            if field == 'output_format' and value not in output_format_choices.choices:
                error = f'Invalid {section} field {field} value {value} provided. Invalid value choice.'
                self.validation_error_messages.append(error)
            if field == 'credential_backend' and value not in backend_choices.choices:
                error = f'Invalid {section} field {field} value {value} provided. Invalid value choice.'
                self.validation_error_messages.append(error)
            if field == 'auto-refresh-profile-cache' and value not in ['true', 'false']:
                error = f'Invalid {section} field {field} value {value} provided. Invalid value choice.'
                self.validation_error_messages.append(error)
            if field == 'auto-refresh-kube-config' and value not in ['true', 'false']:
                error = f'Invalid {section} field {field} value {value} provided. Invalid value choice.'
                self.validation_error_messages.append(error)
            if field == 'default_tenant':
                tenant_aliases_from_sections = [extract_tenant(t) for t in self.config if t.startswith('tenant-')]
                if value not in tenant_aliases_from_sections:
                    error = f'Invalid {section} field {field} value {value} provided. Tenant not found.'
                    self.validation_error_messages.append(error)
            if field == 'ca_bundle':
                ca_bundle_file_path = Path(value).expanduser()
                if not Path.is_file(ca_bundle_file_path):
                    error = f'Invalid {field} file {ca_bundle_file_path}. File does not exist.'
                    self.validation_error_messages.append(error)

    def validate_profile_aliases(self, section, fields):
        for field, value in fields.items():
            if len(profile_split(value)) not in [2, 3]:
                error = (
                    f'Invalid {section} field {field} value {value} provided. Value must be 2 or 3 parts '
                    'separated by a /'
                )
                self.validation_error_messages.append(error)

    def validate_aws(self, section, fields):
        for field, value in fields.items():
            if field not in aws_fields:
                self.validation_error_messages.append(f'Invalid {section} field {field} provided.')
            if field == 'default_checkout_mode' and value not in mode_choices.choices:
                error = f'Invalid {section} field {field} value {value} provided. Invalid value choice.'
                self.validation_error_messages.append(error)

    def validate_gcp(self, section, fields):
        for field, value in fields.items():
            if field not in gcp_fields:
                self.validation_error_messages.append(f'Invalid {section} field {field} provided.')

    def validate_tenant(self, section, fields):
        for field, value in fields.items():
            if field not in tenant_fields:
                self.validation_error_messages.append(f'Invalid {section} field {field} provided.')
            if field == 'name':
                try:
                    Britive.parse_tenant(value)
                except Exception as e:
                    raise click.ClickException(f'Error validating tenant name: {str(e)}')
            if field == 'output_format' and value not in output_format_choices.choices:
                error = f'Invalid {section} field {field} value {value} provided. Invalid value choice.'
                self.validation_error_messages.append(error)

    def auto_refresh_profile_cache(self):
        self.load()
        value = self.config.get('global', {}).get('auto-refresh-profile-cache', 'false')
        if value == 'true':
            return True
        return False

    def auto_refresh_kube_config(self):
        self.load()
        value = self.config.get('global', {}).get('auto-refresh-kube-config', 'false')
        if value == 'true':
            return True
        return False



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options


@click.command()
@build_britive
@britive_options(names='tenant,token,silent,passphrase,federation_provider')
def user(ctx, tenant, token, silent, passphrase, federation_provider):
    """Print details about the authenticated identity."""
    ctx.obj.britive.user()



import click

# eval example: eval $(pybritive checkout test -m env)

ssh_push_public_key_choices = click.Choice(
    [
        'ec2-instance-connect',
        'os-login',
        'instance-metadata',
        'default'  # used in flag_value to make the option look like a flag - each CSP can determine what default means
    ],
    case_sensitive=False
)



import click


option = click.option(
    '--passphrase', '-p',
    help='The passphrase to use for encrypting credentials.',
    envvar='PYBRITIVE_ENCRYPTED_CREDENTIAL_PASSPHRASE',
    show_envvar=True,
    show_default=True
)



import os
from pathlib import Path
import conftest


def read_config():
    with open(f'{conftest.home()}/.britive/pybritive.config', 'r', encoding='utf-8') as f:
        return f.read()


def write_npm_config(simple=True):
    home = conftest.home()
    local_home = os.getenv('PYBRITIVE_HOME_DIR')
    unit = os.getenv('PYBRITIVE_UNIT_TESTING')
    if home and not unit:
        path = Path(Path(local_home) / '.britive' / 'config')

        tenant = os.getenv('PYBRITIVE_TEST_TENANT')
        contents = [f'tenantURL = "https://{tenant}.britive-app.com"', 'output_format = "Table"']

        alias = os.getenv('PYBRITIVE_NPM_IMPORT_PROFILE_ALIAS_VALUE')
        if not simple and alias:
            contents = [
                f'tenantURL = "https://{tenant}.britive-app.com"',
                'output_format = "Table"',
                '',
                '[envProfileMap]',
                f'testalias = "{alias}"',
            ]

        path.write_text('\n'.join(contents), encoding='utf-8')


def common_asserts(result, substring: list = None, exit_code: int = 0):
    assert result.exit_code == exit_code
    if substring:
        if isinstance(substring, str):
            substring = [substring]
        data = read_config()
        for sub in substring:
            assert sub in data


def test_configure_tenant_via_flags_no_alias(runner, cli):
    result = runner.invoke(cli, 'configure tenant -t pybritivetest1.dev -f yaml'.split(' '))
    common_asserts(result, substring='[tenant-pybritivetest1.dev]')


def test_configure_tenant_via_flags_no_alias_no_format(runner, cli):
    result = runner.invoke(cli, 'configure tenant -t pybritivetest2.dev'.split(' '))
    common_asserts(result, substring='[tenant-pybritivetest2.dev]')


def test_configure_tenant_via_flags_yes_alias(runner, cli):
    result = runner.invoke(cli, 'configure tenant -t pybritivetest1.dev -f yaml -a testalias1'.split(' '))
    common_asserts(result, substring='[tenant-testalias1]')


def test_configure_tenant_via_flags_yes_alias_no_format(runner, cli):
    result = runner.invoke(cli, 'configure tenant -t pybritivetest2.dev -a testalias2'.split(' '))
    common_asserts(result, substring='[tenant-testalias2]')


def test_configure_tenant_via_prompt_no_alias(runner, cli):
    result = runner.invoke(cli, 'configure tenant'.split(' '), input='pybritivetest3.dev\n\njson\n')
    common_asserts(result, substring='[tenant-pybritivetest3.dev]')


def test_configure_tenant_via_prompt_no_alias_no_format(runner, cli):
    result = runner.invoke(cli, 'configure tenant'.split(' '), input='pybritivetest4.dev\n\n\n')
    common_asserts(result, substring='[tenant-pybritivetest4.dev]')


def test_configure_tenant_via_prompt_yes_alias(runner, cli):
    result = runner.invoke(cli, ['configure', 'tenant'], input='pybritivetest3.dev\ntestalias3\nyaml\n')
    common_asserts(result, substring='[tenant-testalias3]')


def test_configure_tenant_via_prompt_yes_alias_no_format(runner, cli):
    result = runner.invoke(cli, ['configure', 'tenant'], input='pybritivetest4.dev\ntestalias4\n\n')
    common_asserts(result, substring='[tenant-testalias4]')


def test_configure_global_via_flags_file_backend(runner, cli):
    result = runner.invoke(cli, 'configure global -t pybritivetest1.dev -f table -b file'.split(' '))
    common_asserts(
        result, substring=['default_tenant=pybritivetest1.dev', 'output_format=table', 'credential_backend=file']
    )


def test_configure_global_via_flags_encrypted_file_backend(runner, cli):
    result = runner.invoke(cli, 'configure global -t pybritivetest2.dev -f yaml -b encrypted-file'.split(' '))
    common_asserts(
        result,
        substring=['default_tenant=pybritivetest2.dev', 'output_format=yaml', 'credential_backend=encrypted-file'],
    )


def test_configure_global_via_prompt_file_backend(runner, cli):
    result = runner.invoke(cli, 'configure global'.split(' '), input='pybritivetest1.dev\ntable-pretty\nfile\n')
    common_asserts(
        result, substring=['default_tenant=pybritivetest1.dev', 'output_format=table-pretty', 'credential_backend=file']
    )


def test_configure_global_via_prompt_encrypted_file_backend(runner, cli):
    result = runner.invoke(cli, 'configure global'.split(' '), input='pybritivetest2.dev\n\nencrypted-file\n')
    common_asserts(
        result,
        substring=['default_tenant=pybritivetest2.dev', 'output_format=json', 'credential_backend=encrypted-file'],
    )


def test_configure_global_with_invalid_format(runner, cli):
    result = runner.invoke(cli, 'configure global -f error -P'.split(' '))
    assert "Invalid value for '--format' / '-f'" in result.output


def test_configure_global_with_invalid_tenant(runner, cli):
    result = runner.invoke(cli, 'configure global -t incorrect'.split(' '))
    assert "Invalid global field default_tenant value incorrect provided. Tenant not found." in result.output


def test_configure_import_simple(runner, cli):
    write_npm_config(True)
    tenant = os.getenv('PYBRITIVE_TEST_TENANT')
    result = runner.invoke(cli, 'configure import'.split(' '))
    common_asserts(result, substring=[tenant])


def test_configure_import_complex(runner, cli):
    write_npm_config(False)
    tenant = os.getenv('PYBRITIVE_TEST_TENANT')
    result = runner.invoke(cli, 'configure import'.split(' '))
    print(result.output)
    assert f'Found tenant {tenant}.' in result.output
    assert 'Found default output format' in result.output
    assert 'Profile aliases exist...will retrieve profile details from the tenant.' in result.output
    assert 'Saved alias testalias to profile' in result.output
    common_asserts(result, substring=[tenant, '[profile-aliases]', 'testalias='])


def test_configure_update_global_invalid_data(runner, cli):
    result = runner.invoke(cli, 'configure update global default_tenant incorrect'.split(' '))
    assert result.exit_code == 1
    assert 'Invalid global field default_tenant value incorrect provided. Tenant not found.' in result.output


def test_configure_update_invalid_section(runner, cli):
    result = runner.invoke(cli, 'configure update test default_tenant incorrect'.split(' '))
    assert result.exit_code == 1
    assert 'Cannot save config file due to invalid data provided.' in result.output


def test_configure_update_global_correct_data(runner, cli):
    result = runner.invoke(cli, 'configure update global output_format table-presto'.split(' '))
    common_asserts(result, substring=['output_format=table-presto'])


def test_configure_update_tenant_correct_data(runner, cli):
    tenant = os.getenv('PYBRITIVE_TEST_TENANT')
    result = runner.invoke(cli, f'configure update tenant-{tenant} name pybritivetest1.dev'.split(' '))
    common_asserts(result, substring=['name=pybritivetest1.dev'])
    # set it back
    runner.invoke(cli, f'configure update tenant-{tenant} name {tenant}'.split(' '))


def test_configure_update_aws_data(runner, cli):
    result = runner.invoke(cli, 'configure update aws default_checkout_mode integrate'.split(' '))
    common_asserts(result, substring=['aws', 'default_checkout_mode=integrate'])



import click


option = click.option(
    '--file', '-F',
    help='Path to a file where the contents of the secret file will be stored. Omitting this option will '
         'result in the file being saved to the current directory with the name provided when the secret '
         'file was initially uploaded. Providing `-` will print the contents of the secret file to stdout.'
)



import click

option = click.option(
    '--blocktime', '-b',
    default=3,
    show_default=True,
    help='Seconds to wait before starting to poll for credentials. If not provided will default to 60 for profiles '
         'that require approval and 3 for profiles without approval.'
)



import click
from ..choices.mode import mode_choices


# as of v1.1.0 not setting a default value here on purpose as the config file now has an
# aws section which provides a default value if the --mode option is omitted
# the default to `json` will occur now in helpers/cloud_credential_printer::CloudCredentialPrinter.__init__
option = click.option(
    '--mode', '-m',
    type=mode_choices,
    show_choices=True,
    help='The way in which the checked out credentials are presented. `integrate` will place the credentials into '
         'the cloud providers local credential file (AWS only). Value `env` can optionally include terminal specific '
         'options for setting environment variables '
         '(example: env-nix for Linux/Mac, env-wincmd for Windows Command Prompt, env-winps for Windows PowerShell).'
         '`gcloudauth` will save the generated key file/credentials to the pybritive config directory and generate a '
         'gcloud auth command which can be directly evaluated. Will default to `json` if not provided.'
)



def get_args():
    from getopt import getopt  # lazy load
    from sys import argv  # lazy load
    options = getopt(argv[1:], 't:T:p:f:P:F:hv', [
        'tenant=',
        'token=',
        'passphrase=',
        'force-renew=',
        'profile=',
        'federation-provider='
        'help',
        'version'
    ])[0]

    args = {
        'tenant': None,
        'token': None,
        'passphrase': None,
        'force_renew': None,
        'profile': None,
        'federation_provider': None
    }

    for opt, arg in options:
        if opt in ('-t', '--tenant'):
            args['tenant'] = arg
        if opt in ('-T', '--token'):
            args['token'] = arg
        if opt in ('-p', '--passphrase'):
            args['passphrase'] = arg
        if opt in ('-f', '--force-renew'):
            args['force_renew'] = int(arg)
        if opt in ('-P', '--profile'):
            args['profile'] = arg
        if opt in ('-F', '--federation-provider'):
            args['federation_provider'] = arg
        if opt in ('-h', '--help'):
            usage()
        if opt in ('-v', '--version'):
            from platform import platform, python_version  # lazy load
            from pkg_resources import get_distribution  # lazy load
            cli_version = get_distribution('pybritive').version
            print(
                f'pybritive: {cli_version} / platform: {platform()} / python: {python_version()}'
            )
            raise SystemExit()

    return args


def usage():
    from sys import argv  # lazy load
    print(
        f'Usage : {argv[0]} --profile <profile> [-t/--tenant, -T/--token, -p/--passphrase, -f/--force-renew, '
        f'-F/--federation-provider]'
    )
    raise SystemExit()


def main():
    args = get_args()
    if not args['profile']:
        print('-P/--profile is required')
        usage()

    creds = None
    if not args['force_renew']:  # if force renew let's defer to that the full package vs. this helper
        from .cache import Cache  # lazy load
        creds = Cache(passphrase=args['passphrase']).get_credentials(
            profile_name=args['profile'],
            mode='awscredentialprocess'
        )
        if creds:
            from datetime import datetime  # lazy load
            expiration = datetime.fromisoformat(creds['expirationTime'].replace('Z', ''))
            now = datetime.utcnow()
            if now > expiration:  # creds have expired so set to none so new one get checked out
                creds = None
            else:  # not importing json library on purpose to keep imports down for speed
                json = '{'
                json += f'"AccessKeyId": "{creds["accessKeyID"]}",'
                json += f'"SecretAccessKey": "{creds["secretAccessKey"]}",'
                json += f'"SessionToken": "{creds["sessionToken"]}",'
                json += f'"Expiration": "{creds["expirationTime"]}",'
                json += '"Version": 1}'
                print(json)
                raise SystemExit()
    if not creds:
        from ..britive_cli import BritiveCli  # lazy load for performance purposes

        b = BritiveCli(
            tenant_name=args['tenant'],
            token=args['token'],
            passphrase=args['passphrase'],
            federation_provider=args['federation_provider'],
            silent=True,
            from_helper_console_script=True
        )
        b.config.get_tenant()  # have to load the config here as that work is generally done
        b.checkout(
            alias=None,
            blocktime=None,
            console=False,
            justification=None,
            mode='awscredentialprocess',
            maxpolltime=None,
            profile=args['profile'],
            passphrase=args['passphrase'],
            force_renew=args['force_renew'],
            aws_credentials_file=None,
            gcloud_key_file=None,
            verbose=None,
            extend=False,
            otp=None,
        )
        raise SystemExit()


if __name__ == '__main__':
    main()



import base64
import os
import uuid

from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


class InvalidPassphraseException(Exception):
    pass


class StringEncryption:
    def __init__(self, passphrase: str = None):
        self.passphrase = passphrase or str(uuid.getnode())  # TODO change?

    @staticmethod
    def _salt():
        return base64.b64encode(os.urandom(32)).decode('utf-8')

    def _key(self, salt: str):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=base64.b64decode(salt.encode()),
            iterations=100000,
            backend=default_backend()
        )
        return base64.urlsafe_b64encode(kdf.derive(self.passphrase.encode()))

    def encrypt(self, plaintext: str) -> str:
        salt = self._salt()
        key = self._key(salt)
        ciphertext = Fernet(key).encrypt(plaintext.encode('utf-8'))
        return f'{base64.b64encode(ciphertext).decode("utf-8")}:{salt}'

    def decrypt(self, ciphertext: str):
        try:
            ciphertext, b64salt = ciphertext.split(':')
            key = self._key(b64salt)
            return Fernet(key).decrypt(base64.b64decode(ciphertext.encode())).decode('utf-8')
        except InvalidToken as e:
            raise InvalidPassphraseException() from e



import json


def test_checkout_json(runner, cli, profile):
    result = runner.invoke(cli, ['checkout', profile, '-m', 'json'])
    assert result.exit_code == 0
    creds = json.loads(result.output)
    assert 'AccessKeyId' in creds.keys()
    assert 'SecretAccessKey' in creds.keys()
    assert 'SessionToken' in creds.keys()
    assert 'Expiration' in creds.keys()


def test_checkout_aws_credential_process(runner, cli, profile):
    result = runner.invoke(cli, ['checkout', profile, '-m', 'awscredentialprocess'])
    assert result.exit_code == 0
    creds = json.loads(result.output)
    assert 'AccessKeyId' in creds.keys()
    assert 'SecretAccessKey' in creds.keys()
    assert 'SessionToken' in creds.keys()
    assert 'Expiration' in creds.keys()
    assert 'Version' in creds.keys()



import os
import typing as t

from click.parser import split_arg_string
from click.shell_completion import add_completion_class
from click.shell_completion import ShellComplete, CompletionItem

# inspired by https://raw.githubusercontent.com/tibortakacs/powershell-argcomplete/master/mat.complete.ps1
_powershell_source = """\
$%(complete_func)s = {
    param($wordToComplete, $commandAst, $cursorPosition)

    # in case of scripts, this object holds the current line after string conversion
    $line = "$commandAst"

    # The behaviour of completion should depend on the trailing spaces in the current line:
    # * "command subcommand " --> TAB --> Completion items parameters/sub-subcommands of "subcommand"
    # * "command subcom" --> TAB --> Completion items to extend "subcom" into matching subcommands.
    # $line never contains the trailing spaces. However, $cursorPosition is the length of the original
    # line (with trailing spaces) in this case. This comparison allows the expected user experience.
    if ($cursorPosition -gt $line.Length) {
        $line = "$line "
    }

    # set environment variables that pybritive completion will use
    New-Item -Path Env: -Name COMP_LINE -Value $line | Out-Null # Current line
    New-Item -Path Env: -Name %(complete_var)s -Value "powershell_complete" | Out-Null

    # call %(prog_name)s and it will inspect env vars and provide completion results
    Invoke-Expression %(prog_name)s -ErrorAction SilentlyContinue | Tee-Object -Var completionResult | Out-Null

    # cleanup environment variables
    Remove-Item Env:\\COMP_LINE | Out-Null
    Remove-Item Env:\\%(complete_var)s | Out-Null

    # get list of completion items
    $items = $completionResult -split '\\r?\\n'
    
    $items | ForEach-Object {"$_ "} # trailing space important as completion is "done"
}

# register tab completion
Register-ArgumentCompleter -Native -CommandName %(prog_name)s -ScriptBlock $%(complete_func)s
"""


@add_completion_class
class PowershellComplete(ShellComplete):
    name = "powershell"
    source_template = _powershell_source

    def get_completion_args(self):
        line = os.environ["COMP_LINE"]
        cwords = split_arg_string(line)
        num_cwords = len(cwords)
        args = cwords[1:num_cwords]

        if line.endswith(' '):
            incomplete = ''
        else:
            try:
                incomplete = args.pop()
            except IndexError:
                incomplete = ''
        return args, incomplete

    def format_completion(self, item: CompletionItem) -> str:
        value = item.value
        if ' ' in value:
            value = f'"{value}"'
        return f"{value}"

    def source_vars(self) -> t.Dict[str, t.Any]:
        """Vars for formatting :attr:`source_template`.
        By default this provides ``complete_func``, ``complete_var``,
        and ``prog_name``.
        """
        return {
            "complete_func": self.func_name[1:],  # remove leading _
            "complete_var": self.complete_var,
            "prog_name": self.prog_name
        }



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options
from ..helpers.profile_argument_decorator import click_smart_profile_argument


@click.command()
@build_britive
@britive_options(names='alias,blocktime,console,justification,otp,mode,maxpolltime,silent,force_renew,aws_credentials_file,'
                       'gcloud_key_file,verbose,extend,profile_type,tenant,token,passphrase,federation_provider')
@click_smart_profile_argument
def checkout(ctx, alias, blocktime, console, justification, otp, mode, maxpolltime, silent, force_renew,
             aws_credentials_file, gcloud_key_file, verbose, extend, profile_type, tenant, token, passphrase,
             federation_provider, profile):
    """Checkout a profile.

    This command takes 1 required argument `PROFILE`. This should be a string representation of the profile
    that should be checked out. Format is `application name/environment name/profile name`.
    """

    # silent will get passed in via @build_britive
    ctx.obj.britive.checkout(
        alias=alias,
        blocktime=blocktime,
        console=console,
        justification=justification,
        otp=otp,
        mode=mode,
        maxpolltime=maxpolltime,
        profile=profile,
        passphrase=passphrase,
        force_renew=force_renew,
        aws_credentials_file=aws_credentials_file,
        gcloud_key_file=gcloud_key_file,
        verbose=verbose,
        extend=extend,
        profile_type=profile_type
    )



def common_asserts(result, substring=None, exit_code=0):
    assert result.exit_code == exit_code
    if isinstance(substring, str):
        substring = [substring]
    for sub in substring:
        assert sub in result.output


def test_api(runner, cli):
    result = runner.invoke(cli, 'api users.list'.split(' '))
    common_asserts(result, ['userId', 'status', 'email', 'identityProvider'])



import click

option = click.option(
    '--force-renew',
    default=None,
    show_default=True,
    type=int,
    help='AWS Programmatic Only - If the credentials are to expire within the specified number of minutes, check in '
         'the profile first and check it out again to get a new set of credentials.'
)



import click
from ..options.tenant import option as tenant
from ..options.token import option as token
from ..options.output_format import option as output_format
from ..options.alias import option as alias
from ..options.blocktime import option as blocktime
from ..options.mode import option as mode
from ..options.maxpolltime import option as maxpolltime
from ..options.version import option as version
from ..options.configure_tenant import option as configure_tenant
from ..options.configure_alias import option as configure_alias
from ..options.configure_prompt import option as configure_prompt
from ..options.justification import option as justification
from ..options.otp import option as otp
from ..options.silent import option as silent
from ..options.console import option as console
from ..options.checked_out import option as checked_out
from ..options.file import option as file
from ..options.configure_backend import option as configure_backend
from ..options.passphrase import option as passphrase
from ..options.force_renew import option as force_renew
from ..options.aws_credentials_file import option as aws_credentials_file
from ..options.federation_provider import option as federation_provider
from ..options.gcloud_key_file import option as gcloud_key_file
from ..options.verbose import option as verbose
from ..options.query import option as query
from ..options.ssh_hostname import option as ssh_hostname
from ..options.ssh_username import option as ssh_username
from ..options.ssh_push_public_key import option as ssh_push_public_key
from ..options.ssh_port import option as ssh_port
from ..options.ssh_key_source import option as ssh_key_source
from ..options.aws_profile import option as aws_profile
from ..options.aws_console_duration import option as aws_console_duration
from ..options.browser import option as browser
from ..options.extend import option as extend
from ..options.profile_type import option as profile_type


options_map = {
    'tenant': tenant,
    'token': token,
    'format': output_format,
    'output_format': output_format,
    'alias': alias,
    'blocktime': blocktime,
    'mode': mode,
    'maxpolltime': maxpolltime,
    'version': version,
    'configure_tenant': configure_tenant,
    'configure_alias': configure_alias,
    'configure_prompt': configure_prompt,
    'justification': justification,
    'otp': otp,
    'silent': silent,
    'console': console,
    'checked_out': checked_out,
    'file': file,
    'configure_backend': configure_backend,
    'passphrase': passphrase,
    'force_renew': force_renew,
    'aws_credentials_file': aws_credentials_file,
    'federation_provider': federation_provider,
    'gcloud_key_file': gcloud_key_file,
    'verbose': verbose,
    'query': query,
    'ssh_hostname': ssh_hostname,
    'ssh_username': ssh_username,
    'ssh_push_public_key': ssh_push_public_key,
    'ssh_port': ssh_port,
    'ssh_key_source': ssh_key_source,
    'aws_profile': aws_profile,
    'aws_console_duration': aws_console_duration,
    'browser': browser,
    'extend': extend,
    'profile_type': profile_type
}


def britive_options(*args, **kwargs):
    def inner(f):
        names = [n.strip() for n in kwargs['names'].split(',')]
        names.reverse()
        for name in names:
            option = options_map.get(name)
            if not name:
                raise click.ClickException(f'Invalid option {name} provided.')
            f = option(f)
        return f
    return inner



import json
import os


class KubernetesExecCredentialProcessor:
    def __init__(self):
        self.api_version = None
        self.profile = None
        self.exec_data = None
        self._parse()

    def _parse(self):
        # parse the information provided by kube exec
        self.exec_data = json.loads(os.environ.get('KUBERNETES_EXEC_INFO', 'null'))

        if not self.exec_data:  # this env var HAS to exist if we are being invoked by k8s kubeconfig exec command
            raise Exception(
                'could not find environment variable KUBERNETES_EXEC_INFO - is this command being run '
                'within a kubeconfig context?'
            )

        self.api_version = self.exec_data.get('apiVersion')

        if not self.api_version:
            raise ValueError('apiVersion not found. Cannot continue.')

        if self.api_version == 'client.authentication.k8s.io/v1alpha1':
            raise ValueError(f'apiVersion {self.api_version} is not supported.')

        if self.api_version not in ['client.authentication.k8s.io/v1', 'client.authentication.k8s.io/v1beta1']:
            raise Exception(f'apiVersion {self.api_version} not accounted for.')

        self.profile = self.exec_data.get('spec', {}).get('cluster', {}).get('config', {}).get('britive-profile')

        if not self.profile:
            raise ValueError(
                'kubeconfig cluster extension named britive-profile not found or no profile value specified'
            )

    def construct_exec_credential(self, credentials: dict):
        response = {
            'kind': 'ExecCredential',
            'apiVersion': self.api_version,
            'spec': {},
            'status': {
                'expirationTimestamp': credentials['expirationTime'],
                'token': credentials['jwt']
            }
        }
        return json.dumps(response)



import csv
from datetime import datetime
from datetime import timedelta
from datetime import timezone
import hashlib
import io
import json
import os
from pathlib import Path
import sys
import uuid
import click
from britive import exceptions
from britive.britive import Britive
import jmespath
import jwt
from tabulate import tabulate
import yaml
from .helpers import cloud_credential_printer as printer
from .helpers.cache import Cache
from .helpers.config import ConfigManager
from .helpers.credentials import FileCredentialManager, EncryptedFileCredentialManager
from .helpers.split import profile_split
from . import __version__

default_table_format = 'fancy_grid'
debug_enabled = os.getenv('PYBRITIVE_DEBUG')
default_browser = os.getenv('PYBRITIVE_BROWSER')


class BritiveCli:
    def __init__(
        self,
        tenant_name: str = None,
        token: str = None,
        silent: bool = False,
        passphrase: str = None,
        federation_provider: str = None,
        from_helper_console_script: bool = False,
    ):
        self.silent = silent
        self.from_helper_console_script = from_helper_console_script
        self.output_format = None
        self.tenant_name = None
        self.tenant_alias = None
        self.token = token
        self.b = None
        self.available_profiles = None
        self.config = ConfigManager(tenant_name=tenant_name, cli=self)
        self.list_separator = '|'
        self.passphrase = passphrase
        self.federation_provider = federation_provider
        self.credential_manager = None
        self.verbose_checkout = False
        self.checkout_progress_previous_message = None
        self.cachable_modes = {
            'awscredentialprocess': {'app_type': 'AWS', 'expiration_jmespath': 'expirationTime'},
            'kube-exec': {'app_type': 'Kubernetes', 'expiration_jmespath': 'expirationTime'},
        }
        self.browser = default_browser
        self.resource_profile_prefix = 'resources/'

    def set_output_format(self, output_format: str):
        self.output_format = self.config.get_output_format(output_format)

    def set_credential_manager(self):
        if self.credential_manager:
            return
        backend = self.config.backend()
        if backend == 'file':
            self.credential_manager = FileCredentialManager(
                tenant_alias=self.tenant_alias,
                tenant_name=self.tenant_name,
                cli=self,
                federation_provider=self.federation_provider,
                browser=self.browser,
            )
        elif backend == 'encrypted-file':
            self.credential_manager = EncryptedFileCredentialManager(
                tenant_alias=self.tenant_alias,
                tenant_name=self.tenant_name,
                cli=self,
                passphrase=self.passphrase,
                federation_provider=self.federation_provider,
                browser=self.browser,
            )
        else:
            raise click.ClickException(f'invalid credential backend {backend}.')

    @staticmethod
    def _extract_field_from_jwt(token: str, field: str, verify: bool = False):
        try:
            return jwt.decode(
                token,
                # validation of the token will occur on the Britive backend
                # so not verifying everything here is okay since we are just
                # trying to extract the token expiration time so we can store
                # it in the ~/.britive/pybritive.credentials[.encrypted] file
                options={'verify_signature': verify, 'verify_aud': verify},
            )[field]
        except Exception:
            return None

    def login(self, explicit: bool = False, browser: str = default_browser):
        # explicit means the user called pybritive login, otherwise it is being implicitly called by something else

        self.browser = browser
        self.tenant_name = self.config.get_tenant()['name']
        self.tenant_alias = self.config.alias
        if explicit and self.token:
            raise click.ClickException('Interactive login unavailable when an API token is provided.')

        # taking a very straightforward approach here...if user provided a token and it doesn't work just exit
        if self.token:  # static token provided or BRITIVE_API_TOKEN set
            try:
                self.b = Britive(tenant=self.tenant_name, token=self.token, query_features=False)
                self.b.my_access.whoami()  # this is what may cause UnauthorizedRequest
            except exceptions.UnauthorizedRequest as e:
                raise click.ClickException('Invalid API token provided.') from e
            except exceptions.InvalidRequest as e:
                if '400 - e1000 - bad request' in str(e).lower():  # this is for SCIM token
                    pass
                else:
                    raise e
        else:  # user is asking for an interactive login or using token stored from an interactive login
            counter = 1
            while True:  # will break after we successfully get logged in or 3 attempts have occurred
                # protect against infinite loop
                if counter > 3:
                    raise Exception('could not login after 3 attempts')

                # attempt login and making an api call to ensure the credentials we have are valid
                try:
                    self.set_credential_manager()
                    token = self.credential_manager.get_token()
                    jti = self._extract_field_from_jwt(token=token, field='jti')
                    self.debug(f'got token jti of {jti} from credential manager')
                    self.b = Britive(tenant=self.tenant_name, token=token, query_features=False)
                    self.b.my_access.whoami()  # this is what may cause UnauthorizedRequest
                    break
                except exceptions.UnauthorizedRequest as e:
                    if '401 - e0000' in str(e).lower():
                        self.debug(f'attempt {counter} of 3 - login failed')
                        self.debug(f'login error message was {str(e)}')

                        # we know the token is invalid since we got that API response
                        # so we don't need to actually logout, just clear the token from
                        # the credentials manager
                        self._cleanup_credentials()
                    else:
                        raise e
                finally:
                    counter += 1

        self._update_sdk_user_agent()
        # if user called `pybritive login` and we should get profiles...do so
        should_get_profiles = any([self.config.auto_refresh_profile_cache(), self.config.auto_refresh_kube_config()])
        if explicit and should_get_profiles:
            self._set_available_profiles()  # will handle calling cache_profiles() and construct_kube_config()

        self._display_banner()

    def _display_banner(self):
        if self.silent:
            return

        if not Cache().banner_expired(tenant=self.tenant_name):  # if banner is not expired yet then nothing to do
            return

        # if we get here then we need to at least grab the banner and see if it has changed
        banner = self.b.banner()
        banner_changed = Cache().save_banner(tenant=self.tenant_name, banner=banner)
        if banner and banner_changed:
            self.print(f'*** {banner.get("messageType", "UNKNOWN")}: {banner.get("message", "<no message>")} ***')

    def _update_sdk_user_agent(self):
        # update the user agent to include the pybritive cli version
        user_agent = self.b.session.headers.get('User-Agent')

        try:
            version = __version__
        except Exception:
            version = 'unknown'

        self.b.session.headers.update({'User-Agent': f'pybritive/{version} {user_agent}'})

    def _cleanup_credentials(self):
        self.set_credential_manager()
        self.credential_manager.delete()

    @staticmethod
    def _is_saml_user(token):
        try:
            username = jwt.decode(
                token,
                # validation of the token will occur on the Britive backend
                # so not verifying everything here is okay since we are just
                # trying to extract the username to determine if they are a
                # SAML user or not
                options={'verify_signature': False, 'verify_aud': False},
            ).get('username', '')

            return username.startswith('SAML')
        except:
            return False

    def logout(self):
        # if dealing with a token there is no concept of logout
        if self.token:
            raise click.ClickException('Logout not available when using an API token.')

        self.tenant_name = self.config.get_tenant()['name']
        self.tenant_alias = self.config.alias
        self.set_credential_manager()

        # let's see if we have credentials for this tenant already
        # if we do we need to invalidate them at the tenant and clean them up on the client side
        # if we don't have valid credentials for the tenant then there is no need to logout
        if self.credential_manager.has_valid_credentials():
            token = self.credential_manager.get_token()

            # keep it as local variable, so we don't mess up anything that may be happening in login
            # if this method is called due to a 401 E0000 error
            b = Britive(tenant=self.tenant_name, token=token, query_features=False)

            params = {}
            if self._is_saml_user(token):
                params['type'] = 'sso'

            b.delete(f'https://{Britive.parse_tenant(self.tenant_name)}/api/auth', params=params)
            self._cleanup_credentials()

    def debug(self, data: object, ignore_silent: bool = False):
        if debug_enabled:
            self.print(data=data, ignore_silent=ignore_silent)

    # will be passed to the britive checkout_by_name progress_func parameter when appropriate
    def checkout_callback_printer(self, message: str):
        if self.silent or not sys.stdout.isatty():
            return
        if message == 'complete':
            click.echo('')
            return

        if self.verbose_checkout:
            if self.checkout_progress_previous_message != message:
                newline = '\n' if self.checkout_progress_previous_message else ''
                self.checkout_progress_previous_message = message
                click.echo(f'{newline}{message} ', nl=False)
            else:
                click.echo('.', nl=False)
        else:
            click.echo('.', nl=False)

    # will take a list of dicts and print to the screen based on the format specified in the config file
    # dict can only be 1 level deep (no nesting) - caller needs to massage the data accordingly
    def print(self, data: object, ignore_silent: bool = False):
        if self.silent and not ignore_silent:
            return

        if isinstance(data, str):  # if we have a string just print it and move on
            click.echo(data)
            return

        if self.output_format == 'json':
            click.echo(json.dumps(data, indent=2, default=str))
        elif self.output_format == 'list-profiles':
            for row in data:
                click.echo(self.list_separator.join([self.escape_profile_element(x) for x in row.values()]))
        elif self.output_format == 'list':
            for row in data:
                if isinstance(row, dict):
                    click.echo(self.list_separator.join([json.dumps(x, default=str) for x in row.values()]))
                elif isinstance(row, list):
                    click.echo(self.list_separator.join([json.dumps(x, default=str) for x in row]))
                else:
                    click.echo(row)
        elif self.output_format == 'csv':
            fields = list(data[0])
            output = io.StringIO()
            writer = csv.DictWriter(output, fieldnames=fields, delimiter=',')
            writer.writeheader()
            writer.writerows(data)
            click.echo(output.getvalue())
        elif self.output_format.startswith('table'):
            if isinstance(data, dict):
                data = [data]
            tablefmt = default_table_format
            split = self.output_format.split('-')
            if len(split) > 1:
                tablefmt = split[1]
            click.echo(tabulate(data, headers='keys', tablefmt=tablefmt))
        elif self.output_format == 'yaml':
            y = yaml.safe_load(json.dumps(data))
            click.echo(yaml.safe_dump(y))
        else:
            raise click.ClickException(f'Invalid output format {self.output_format} provided.')

    def user(self):
        self.login()
        username = self.b.my_access.whoami()['username']
        alias = self.tenant_alias
        output = f'{username} @ {self.tenant_name}'
        if alias != self.tenant_name:
            output += f' (alias: {alias})'
        self.print(output, ignore_silent=True)

    def list_secrets(self):
        self.login()
        self.print(self.b.my_secrets.list(), ignore_silent=True)

    def list_approvals(self):
        self.login()
        approvals = []
        for approval in self.b.my_access.list_approvals():
            approval.pop('resource', None)
            approval.pop('consumer', None)
            approval.pop('timeToApprove', None)
            approval.pop('validFor', None)
            approval.pop('action', None)
            approval.pop('approvers', None)
            approval.pop('expirationTimeApproval', None)
            approval.pop('updatedAt', None)
            approval.pop('actionBy', None)
            approval.pop('validForInDays', None)
            approvals.append(approval)

        approvals = sorted(approvals, key=lambda x: x['createdAt'])
        approvals.reverse()
        self.print(approvals, ignore_silent=True)

    def list_resources(self):
        self.login()
        found_resource_names = []
        resources = []
        for item in self.b.my_resources.list_profiles():
            name = item['resourceName']
            if name not in found_resource_names:
                resources.append(
                    {
                        'resourceId': item['resourceId'],
                        'resourceName': name,
                        'resourceLabels': item['resourceLabels']
                    }
                )
                found_resource_names.append(name)
        self.print(resources, ignore_silent=True)

    def list_profiles(self, checked_out: bool = False, profile_type: str = None):
        self.login()
        self._set_available_profiles(profile_type=profile_type)
        data = []
        checked_out_profiles = {}
        if checked_out:  # only make this call if we have to
            now = datetime.utcnow()
            for p in self.b.my_access.list_checked_out_profiles():
                expiration_str = p['expiration']
                expiration_timestamp = datetime.fromisoformat(expiration_str.replace('Z', ''))
                seconds_until_expiration = int((expiration_timestamp - now).total_seconds())
                key = f'{p["papId"]}-{p["environmentId"]}'
                checked_out_profiles[key] = {
                    'expiration': expiration_str,
                    'expires_in_seconds': seconds_until_expiration,
                }

        for profile in self.available_profiles:
            key = f'{profile["profile_id"]}-{profile["env_id"]}'
            profile_is_checked_out = key in checked_out_profiles
            if not checked_out or profile_is_checked_out:
                row = {
                    'Application': profile['app_name'] or 'Resources',
                    'Environment': profile['env_name'],
                    'Profile': profile['profile_name'],
                    'Description': profile['profile_description'] or 'Resource',
                    'Type': profile['app_type'],
                }

                if profile_is_checked_out:
                    row['Expiration'] = checked_out_profiles[key]['expiration']
                    total_seconds = checked_out_profiles[key]['expires_in_seconds']

                    hours, remainder = divmod(total_seconds, 3600)
                    minutes, seconds = divmod(remainder, 60)
                    time_format = f'{hours:02d}:{minutes:02d}:{seconds:02d}'
                    row['TimeRemaining'] = time_format
                    row['TimeRemainingSeconds'] = total_seconds

                if self.output_format == 'list':
                    self.list_separator = '/'
                    row.pop('Description', None)
                    row.pop('Type', None)
                    row.pop('TimeRemaining', None)
                    row.pop('TimeRemainingSeconds', None)
                    row.pop('Expiration', None)
                    if profile['2_part_profile_format_allowed']:
                        row.pop('Environment', None)
                elif self.output_format == 'json':
                    row['Name'] = f"{row['Application']}/{row['Environment']}/{row['Profile']}"

                data.append(row)

        # set special list output if needed
        if self.output_format == 'list':
            self.output_format = 'list-profiles'

        self.print(data, ignore_silent=True)

        # and set it back
        if self.output_format == 'list-profiles':
            self.output_format = 'list'

    def list_applications(self):
        self.login()
        self._set_available_profiles()
        keys = ['app_name', 'app_type', 'app_description']
        apps = []
        for profile in self.available_profiles:
            apps.append({k: v for k, v in profile.items() if k in keys})
        apps = [dict(t) for t in {tuple(d.items()) for d in apps}]  # de-dup
        data = []
        for app in apps:
            row = {
                'Application': app['app_name'],
                'Type': app['app_type'],
                'Description': app['app_description'],
            }
            data.append(row)
        self.print(data, ignore_silent=True)

    def list_environments(self):
        self.login()
        self._set_available_profiles()
        envs = []
        keys = ['app_name', 'app_type', 'env_name', 'env_description']
        for profile in self.available_profiles:
            envs.append({k: v for k, v in profile.items() if k in keys})
        envs = [dict(t) for t in {tuple(d.items()) for d in envs}]  # de-dup

        data = []
        for env in envs:
            row = {
                'Application': env['app_name'],
                'Environment': env['env_name'],
                'Description': env['env_description'],
                'Type': env['app_type'],
            }
            data.append(row)
        self.print(data, ignore_silent=True)

    def _set_available_profiles(self, from_cache_command=False, profile_type: str = None):
        if not self.available_profiles:
            data = []
            if not profile_type or profile_type == 'my-access':
                for app in self.b.my_access.list_profiles():
                    for profile in app.get('profiles', []):
                        for env in profile.get('environments', []):
                            row = {
                                'app_name': app['appName'],
                                'app_id': app['appContainerId'],
                                'app_type': app['catalogAppName'],
                                'app_description': app['appDescription'],
                                'env_name': env['environmentName'],
                                'env_id': env['environmentId'],
                                'env_short_name':  env['alternateEnvironmentName'],
                                'env_description': env['environmentDescription'],
                                'profile_name': profile['profileName'],
                                'profile_id': profile['profileId'],
                                'profile_allows_console': profile['consoleAccess'],
                                'profile_allows_programmatic': profile['programmaticAccess'],
                                'profile_description': profile['profileDescription'],
                                '2_part_profile_format_allowed': app['requiresHierarchicalModel'],
                                'env_properties': env.get('profileEnvironmentProperties', {})
                            }
                            data.append(row)
            if not profile_type or profile_type == 'my-resources':
                for item in self.b.my_resources.list_profiles():
                    row = {
                        'app_name': None,
                        'app_id': None,
                        'app_type': 'Resources',
                        'app_description': None,
                        'env_name': item['resourceName'],
                        'env_id': item['resourceId'],
                        'env_short_name': item['resourceName'],
                        'env_description': None,
                        'profile_name': item['profileName'],
                        'profile_id': item['profileId'],
                        'profile_allows_console': False,
                        'profile_allows_programmatic': True,
                        'profile_description': None,
                        '2_part_profile_format_allowed': False,
                        'env_properties': item.get('resourceLabels', {})
                    }
                    data.append(row)
            self.available_profiles = data
            if not from_cache_command and self.config.auto_refresh_profile_cache():
                self.cache_profiles()
            if not from_cache_command and self.config.auto_refresh_kube_config():
                self.construct_kube_config()

    def construct_kube_config(self, from_cache_command=False):
        if self.from_helper_console_script:
            return

        if from_cache_command:
            self.login()
            self._set_available_profiles(from_cache_command=from_cache_command)

        profiles = []
        for p in self.available_profiles:
            if p['app_type'].lower() == 'kubernetes':
                props = p['env_properties']
                url = props.get('apiServerUrl')
                cert = props.get('certificateAuthorityData')
                if props and all([url, cert]):
                    profiles.append(
                        {
                            'app': p['app_name'],
                            'env': p['env_name'],
                            'profile': p['profile_name'],
                            'url': url,
                            'cert': cert,
                        }
                    )

        try:
            from .helpers.kube_config_builder import build_kube_config  # lazy import as not everyone will want this

            build_kube_config(
                profiles=profiles, config=self.config, username=self.b.my_access.whoami()['username'], cli=self
            )
        except Exception as e:  # do NOT fail the CLI invocation because of this
            self.print(f'error auto-generating the Britive managed kube config file: {str(e)}')

    def _get_app_type(self, application_id):
        self._set_available_profiles()
        for profile in self.available_profiles:
            if profile['app_id'] == application_id:
                return profile['app_type']
        raise click.ClickException(f'Application {application_id} not found')

    def __get_cloud_credential_printer(
        self,
        app_type,
        console,
        mode,
        profile,
        silent,
        credentials,
        aws_credentials_file,
        gcloud_key_file,
        k8s_processor,
    ):
        if app_type in ['AWS', 'AWS Standalone']:
            return printer.AwsCloudCredentialPrinter(
                console=console,
                mode=mode or self.config.aws_default_checkout_mode(),  # handle the aws default_checkout_mode here
                profile=profile,
                credentials=credentials,
                silent=silent,
                cli=self,
                aws_credentials_file=aws_credentials_file,
            )
        if app_type in ['Azure']:
            return printer.AzureCloudCredentialPrinter(
                console=console, mode=mode, profile=profile, credentials=credentials, silent=silent, cli=self
            )
        if app_type in ['GCP']:
            return printer.GcpCloudCredentialPrinter(
                console=console,
                mode=mode,
                profile=profile,
                credentials=credentials,
                silent=silent,
                cli=self,
                gcloud_key_file=gcloud_key_file,
            )
        if app_type in ['Kubernetes']:
            return printer.KubernetesCredentialPrinter(
                console=console,
                mode=mode,
                profile=profile,
                credentials=credentials,
                silent=silent,
                cli=self,
                k8s_processor=k8s_processor,
            )
        if app_type in ['OpenShift']:
            return printer.OpenShiftCredentialPrinter(
                console=console,
                mode=mode,
                profile=profile,
                credentials=credentials,
                silent=silent,
                cli=self
            )
        if app_type in ['Resources']:
            return printer.ResourcesCredentialPrinter(
                profile=profile,
                credentials=credentials,
                silent=silent,
                cli=self
            )
        return printer.GenericCloudCredentialPrinter(
            console=console,
            mode=mode,
            profile=profile,
            credentials=credentials,
            silent=silent,
            cli=self,
        )
    def _resource_checkin(self, profile):
        resource_name, profile_name = self._split_resource_profile_into_parts(profile=profile)
        self.login()
        self.b.my_resources.checkin_by_name(
            profile_name=profile_name,
            resource_name=resource_name
        )

    def _access_checkin(self, profile, console):
        self.login()
        self._set_available_profiles()
        parts = self._split_profile_into_parts(profile)

        ids = self._convert_names_to_ids(
            profile_name=parts['profile'], environment_name=parts['env'], application_name=parts['app']
        )

        access_type = 'CONSOLE' if console else 'PROGRAMMATIC'

        transaction_id = None
        application_type = None
        for checked_out_profile in self.b.my_access.list_checked_out_profiles():
            same_env = checked_out_profile['environmentId'] == ids['environment_id']
            same_profile = checked_out_profile['papId'] == ids['profile_id']
            same_access_type = checked_out_profile['accessType'] == access_type
            if all([same_env, same_profile, same_access_type]):
                transaction_id = checked_out_profile['transactionId']

                for available_profile in self.available_profiles:
                    same_env_2 = checked_out_profile['environmentId'] == available_profile['env_id']
                    same_profile_2 = checked_out_profile['papId'] == available_profile['profile_id']
                    if all([same_env_2, same_profile_2, access_type == 'PROGRAMMATIC']):
                        application_type = available_profile['app_type'].lower()
                        break
                break
        if not transaction_id:
            raise ValueError('no checked out profile found for the given profile')

        self.b.my_access.checkin(transaction_id=transaction_id)

        if application_type in ['aws', 'aws standalone']:
            self.clear_cached_aws_credentials(profile)
        if application_type in ['gcp']:
            self.clear_gcloud_auth_key_files(profile=profile)

    def checkin(self, profile, console, profile_type: str = 'my-access'):
        if self._profile_is_for_resource(profile=profile, profile_type=profile_type):
            self._resource_checkin(profile=profile)
        else:
            self._access_checkin(profile=profile, console=console)

    def _checkout(self, profile_name, env_name, app_name, programmatic, blocktime, maxpolltime, justification, otp):
        try:
            self.login()

            ids = self._convert_names_to_ids(
                profile_name=profile_name, environment_name=env_name, application_name=app_name
            )

            return self.b.my_access.checkout(
                profile_id=ids['profile_id'],
                environment_id=ids['environment_id'],
                programmatic=programmatic,
                include_credentials=True,
                wait_time=blocktime,
                max_wait_time=maxpolltime,
                justification=justification,
                otp=otp,
                progress_func=self.checkout_callback_printer,  # callback will handle silent, isatty, etc.
            )
        except exceptions.ApprovalRequiredButNoJustificationProvided as e:
            raise click.ClickException('approval required and no justification provided.') from e
        except ValueError as e:
            raise click.BadParameter(str(e))
        except Exception as e:
            if 'programmatic access is not enabled' in str(e).lower():
                # attempt to automatically checkout console access instead
                # this is a cli only feature - not available in the sdk
                self.print('no programmatic access available - checking out console access instead')
                return self._checkout(
                    profile_name, env_name, app_name, False, blocktime, maxpolltime, justification, otp
                )
            raise e

    @staticmethod
    def _should_check_force_renew(app, force_renew, console):
        return app in ['AWS', 'AWS Standalone'] and force_renew and not console

    def _split_profile_into_parts(self, profile):
        profile_real = self.config.profile_aliases.get(profile.lower(), profile)
        parts = profile_split(profile_real)
        if len(parts) == 2:  # handle shortcut for profiles where the app and environment name are the same
            parts = [parts[0], parts[0], parts[1]]
        if len(parts) != 3:
            raise click.ClickException('Provided profile string does not have the required parts.')
        parts_dict = {'app': parts[0], 'env': parts[1], 'profile': parts[2]}
        return parts_dict

    def _extend_checkout(self, profile, console):
        self.login()
        parts = self._split_profile_into_parts(profile)
        self.b.my_access.extend_checkout_by_name(
            profile_name=parts['profile'],
            environment_name=parts['env'],
            application_name=parts['app'],
            programmatic=not console,
        )

    def _save_alias(self, alias, profile):
        if alias:
            self.config.save_profile_alias(alias=alias, profile=profile.lower())

    def _split_resource_profile_into_parts(self, profile):
        real_profile_name = self.config.profile_aliases.get(profile.lower(), profile).lower()
        if real_profile_name.startswith(self.resource_profile_prefix):
            real_profile_name = real_profile_name.replace(self.resource_profile_prefix, '')
        return real_profile_name.split('/')

    def _profile_is_for_resource(self, profile, profile_type):
        if profile_type == 'my-resources':
            return True
        real_profile_name = self.config.profile_aliases.get(profile.lower(), profile).lower()
        return real_profile_name.startswith(f'{self.resource_profile_prefix}')

    def _resource_checkout(self, blocktime, justification, maxpolltime, profile):
        self.login()
        resource_name, profile_name = self._split_resource_profile_into_parts(profile=profile)
        response = self.b.my_resources.checkout_by_name(
            resource_name=resource_name,
            profile_name=profile_name,
            include_credentials=True,
            justification=justification,
            wait_time=blocktime,
            max_wait_time=maxpolltime,
            progress_func=self.checkout_callback_printer  # callback will handle silent, isatty, etc.
        )
        return response['credentials']

    def _access_checkout(self, alias, blocktime, console, justification, otp, mode, maxpolltime, profile, passphrase,
                         force_renew, verbose, extend):
        # handle this special use case and quit
        if extend:
            self._extend_checkout(profile, console)
            return

        credentials = None
        app_type = None
        cached_credentials_found = False
        k8s_processor = None
        self.verbose_checkout = verbose

        # handle kube-exec since the profile is actually going to be passed in via another method
        # and perform some basic validation so we don't waste time performing a checkout when we
        # will not be able to return a response back to kubectl via the exec command
        if mode == 'kube-exec':
            from .helpers.k8s_exec_credential_builder import KubernetesExecCredentialProcessor

            k8s_processor = KubernetesExecCredentialProcessor()

        # these 3 modes implicitly say that console access should be checked out without having to provide
        # the --console flag
        if mode and (mode == 'console' or mode.startswith('browser') or mode.startswith('os-')):
            console = True
            if mode.startswith('browser'):
                self.browser = mode.replace('browser-', '')
            else:
                self.browser = default_browser

        self._validate_justification(justification)

        if mode in self.cachable_modes:
            self.silent = True  # CANNOT output anything other than the expected JSON
            # we need to check the cache for the credentials first and then check to see if they are expired
            # if not simply return those credentials, if they are expired, continue to do an actual checkout
            app_type = self.cachable_modes[mode]['app_type']
            credentials = Cache(passphrase=passphrase).get_credentials(profile_name=alias or profile, mode=mode)
            if credentials:
                expiration_timestamp_str = jmespath.search(
                    expression=self.cachable_modes[mode]['expiration_jmespath'], data=credentials
                ).replace('Z', '')
                expires = datetime.fromisoformat(expiration_timestamp_str)
                now = datetime.utcnow()
                if now >= expires:  # check to ensure the credentials are still valid, if not, set to None and get new
                    credentials = None
                else:
                    cached_credentials_found = True

        parts = self._split_profile_into_parts(profile)

        # create this params once so we can use it multiple places
        params = {
            'profile_name': parts['profile'],
            'env_name': parts['env'],
            'app_name': parts['app'],
            'programmatic': not console,
            'blocktime': blocktime,
            'maxpolltime': maxpolltime,
            'justification': justification,
            'otp': otp,
        }

        if not cached_credentials_found:  # nothing found in cache, cache is expired, or not a cachable mode
            response = self._checkout(**params)
            app_type = self._get_app_type(response['appContainerId'])
            credentials = response['credentials']

        # this handles the --force-renew flag
        # lets check to see if we should checkin this profile first and check it out again
        if self._should_check_force_renew(app_type, force_renew, console):
            expiration = datetime.fromisoformat(credentials['expirationTime'].replace('Z', ''))
            now = datetime.utcnow()
            diff = (expiration - now).total_seconds() / 60.0
            if diff < force_renew:  # time to checkin the profile so we can refresh creds
                self.print('checking in the profile to get renewed credentials....standby')
                self.checkin(profile=profile, console=console)
                response = self._checkout(**params)
                cached_credentials_found = False  # need to write new creds to cache
                credentials = response['credentials']

        if mode in self.cachable_modes and not cached_credentials_found:
            Cache(passphrase=passphrase).save_credentials(
                profile_name=alias or profile, credentials=credentials, mode=mode
            )
        return app_type, credentials, k8s_processor

    def checkout(self, alias, blocktime, console, justification, otp, mode, maxpolltime, profile, passphrase,
                 force_renew, aws_credentials_file, gcloud_key_file, verbose, extend, profile_type: str = 'my-access'):
        if self._profile_is_for_resource(profile=profile, profile_type=profile_type):
            app_type = 'Resources'
            k8s_processor = None
            credentials = self._resource_checkout(
                blocktime=blocktime,
                justification=justification,
                maxpolltime=maxpolltime,
                profile=profile
            )
        else:
            app_type, credentials, k8s_processor = self._access_checkout(
                alias=alias,
                blocktime=blocktime,
                console=console,
                justification=justification,
                otp=otp,
                mode=mode,
                maxpolltime=maxpolltime,
                profile=profile,
                passphrase=passphrase,
                force_renew=force_renew,
                verbose=verbose,
                extend=extend
            )

        # do this down here, so we know that the profile is valid and a checkout was successful
        self._save_alias(alias=alias, profile=profile)

        self.__get_cloud_credential_printer(
            app_type,
            console,
            mode,
            alias or profile,
            self.silent,
            credentials,
            aws_credentials_file,
            gcloud_key_file,
            k8s_processor,
        ).print()

    def import_existing_npm_config(self):
        profile_aliases = self.config.import_global_npm_config()

        if len(profile_aliases) == 0:
            return
        self.print('')
        self.print('Profile aliases exist...will retrieve profile details from the tenant.')
        self.print('')

        self.login()
        self._set_available_profiles()
        self.print('')
        for alias, ids in profile_aliases.items():
            if '/' in alias:  # no need to import the aliases that aren't really aliases
                continue
            app, env, profile = ids.split('/')[:3]
            for p in self.available_profiles:
                if p['app_id'] == app and p['env_id'] == env and p['profile_id'] == profile:
                    profile_str = f"{p['app_name']}/{p['env_name']}/{p['profile_name']}"
                    self.config.save_profile_alias(alias, profile_str)
                    self.print(f'Saved alias {alias} to profile {profile_str}')

    def configure_tenant(self, tenant, alias, output_format):
        self.config.save_tenant(tenant=tenant, alias=alias, output_format=output_format)

    def configure_global(self, default_tenant_name, output_format, backend):
        self.config.save_global(default_tenant_name=default_tenant_name, output_format=output_format, backend=backend)

    def viewsecret(self, path, blocktime, justification, maxpolltime):
        self._validate_justification(justification)
        self.login()

        try:
            value = self.b.my_secrets.view(
                path=path, justification=justification, wait_time=blocktime, max_wait_time=maxpolltime
            )
        except exceptions.AccessDenied as e:
            raise click.ClickException('user does not have access to the secret.') from e
        except exceptions.ApprovalRequiredButNoJustificationProvided as e:
            raise click.ClickException('approval required and no justification provided.') from e

        # handle the generic note template type for a better UX
        if len(value) == 1 and 'Note' in value:
            value = value['Note']

        # if the value can be converted from JSON to python dict, do it
        try:
            value = json.loads(value)
        except json.JSONDecodeError:
            pass
        except TypeError:
            pass

        # and finally print the secret data
        self.print(value, ignore_silent=True)

    def downloadsecret(self, path, blocktime, justification, maxpolltime, file):
        self._validate_justification(justification)
        self.login()

        try:
            response = self.b.my_secrets.download(
                path=path, justification=justification, wait_time=blocktime, max_wait_time=maxpolltime
            )
        except exceptions.AccessDenied as e:
            raise click.ClickException('user does not have access to the secret.') from e
        except exceptions.ApprovalRequiredButNoJustificationProvided as e:
            raise click.ClickException('approval required and no justification provided.') from e

        filename_from_secret = response['filename']
        content = response['content_bytes']

        if file == '-':
            try:
                self.print(content.decode('utf-8'), ignore_silent=True)
            except UnicodeDecodeError as e:
                raise click.ClickException(
                    'Secret file contents cannot be decoded to utf-8. Save the contents of the file to disk instead.'
                ) from e
            return

        filename = file or filename_from_secret
        path = str(Path(filename).absolute())
        with open(path, 'wb') as f:
            f.write(content)
        self.print(f'wrote contents of secret file to {path}')

    def cache_profiles(self, from_cache_command=False):
        if self.from_helper_console_script:
            return
        profiles = []

        if from_cache_command:
            self.login()
            self._set_available_profiles(from_cache_command=from_cache_command)

        for p in self.available_profiles:
            profile = self.escape_profile_element(p['app_name'])
            profile += '/'

            if not p['2_part_profile_format_allowed']:
                profile += self.escape_profile_element(p['env_name'])
                profile += '/'

            profile += self.escape_profile_element(p['profile_name'])
            profiles.append(profile)
        Cache().save_profiles(profiles)

    @staticmethod
    def escape_profile_element(element):
        if element is None:
            element = 'resources'
        return element.replace('/', '\\/')

    @staticmethod
    def cache_clear():
        Cache().clear()
        Cache().clear_kubeconfig()

    @staticmethod
    def clear_kubeconfig():
        Cache().clear_kubeconfig()

    def configure_update(self, section, field, value):
        self.config.update(section=section, field=field, value=value)

    def request_submit(self, profile, justification):
        self._validate_justification(justification)
        self.login()
        parts = self._split_profile_into_parts(profile)

        ids = self._convert_names_to_ids(
            profile_name=parts['profile'], environment_name=parts['env'], application_name=parts['app']
        )

        self.b.my_access.request_approval(
            profile_id=ids['profile_id'],
            environment_id=ids['environment_id'],
            block_until_disposition=False,
            justification=justification,
        )

    def request_withdraw(self, profile):
        self.login()
        parts = self._split_profile_into_parts(profile)

        ids = self._convert_names_to_ids(
            profile_name=parts['profile'], environment_name=parts['env'], application_name=parts['app']
        )

        self.b.my_access.withdraw_approval_request(profile_id=ids['profile_id'], environment_id=ids['environment_id'])

    @staticmethod
    def build_gcloud_key_file_for_gcloudauthexec(profile: str):
        profile_hash = hashlib.sha256(string=profile.encode('utf-8')).hexdigest()
        return f'gcloudauthexec-{profile_hash}.json'

    def clear_gcloud_auth_key_files(self, profile: str = None):
        if profile:  # we want to attempt a gcloud cli command
            import subprocess  # lazy load as this will not always be needed

            # build the path to the key file in question
            key_file = self.build_gcloud_key_file_for_gcloudauthexec(profile=profile)
            path = Path(self.config.gcloud_key_file_path) / key_file

            if path.exists():  # we have a valid gcloudauthexec key file, so we know there was a checkout with this mode
                try:
                    with open(str(path), 'r', encoding='utf-8') as f:
                        credentials = json.loads(f.read())
                    commands = ['gcloud', 'auth', 'revoke', credentials['client_email'], '--verbosity=error']
                    self.debug(' '.join(commands))
                    subprocess.run(commands, check=True)

                    gcloud_default_account = self.config.gcloud_default_account()

                    if gcloud_default_account:
                        commands = [
                            'gcloud',
                            'config',
                            'set',
                            'account',
                            gcloud_default_account,  # no need for "" here as subprocess will properly escape
                            '--verbosity=error',
                        ]
                        self.debug(' '.join(commands))
                        subprocess.run(commands, check=True)
                except Exception as e:
                    self.print(f'could not reset gcloud CLI active account due to issue: {str(e)}')
        self.config.clear_gcloud_auth_key_files(profile=profile)

    def api(self, method, parameters: dict, query=None):
        self.login()

        # clean up parameters - need to load json as dict if json string is provided and handle file inputs
        computed_parameters = {}
        open_file_keys = []
        try:
            for key, value in parameters.items():
                computed_key = key.replace('-', '_')
                computed_value = value

                if value.lower() == 'none':
                    computed_value = None

                if value.startswith('file://'):
                    filepath = value.replace('file://', '')
                    path = Path(filepath)
                    with open(str(path), 'r', encoding='utf-8') as f:
                        computed_value = f.read().strip()

                if value.startswith('fileb://'):
                    filepath = value.replace('fileb://', '')
                    path = Path(filepath)
                    computed_value = open(str(path), 'rb')
                    open_file_keys.append(computed_key)

                try:
                    computed_parameters[computed_key] = json.loads(computed_value)
                except json.JSONDecodeError:
                    computed_parameters[computed_key] = computed_value
                except Exception:  # not sure what else we would do so just default to the value provided
                    computed_parameters[computed_key] = computed_value
        except AttributeError as e:
            raise click.ClickException(f'invalid parameters {parameters} provided.') from e

        # determine the sdk method we need to execute, starting at the base Britive class
        func = self.b
        try:
            for m in method.split('.'):
                func = getattr(func, m)
        except Exception as e:
            raise click.ClickException(f'invalid method {method} provided.') from e

        # execute the method with the computed parameters
        response = func(**computed_parameters)

        # close any files we opened due to fileb:// prefix
        for key in open_file_keys:
            try:
                computed_parameters[key].close()
            except Exception:
                pass

        # output the response, optionally filtering based on provided jmespath query/search
        self.print(jmespath.search(query, response) if query else response, ignore_silent=True)

    # yes - this method exits in b.my_access as _get_profile_and_environment_ids_given_names
    # but we are doing additional business logic here to enhance the cli experience so there is
    # a need to duplicate some of this logic (although we are doing additional work here so the logic is
    # not 100% duplicated)
    def _convert_names_to_ids(self, profile_name: str, environment_name: str, application_name: str) -> dict:
        # set the available profiles if this is not already done
        self._set_available_profiles()

        # do some sanitization just in case
        profile_name = profile_name.lower().strip()
        environment_name = environment_name.lower().strip()
        application_name = application_name.lower().strip()

        found_profiles = {}

        # collect relevant profile/environment combinations to which the identity is entitled
        for profile in self.available_profiles:
            if profile['app_name'].lower() != application_name:  # kick out all the unmatched applications
                continue
            if profile['profile_name'].lower() != profile_name:  # kick out the unmatched profiles
                continue

            # if we get here we know we are on a record that is the right app and right profile

            found_profile_id = profile['profile_id']

            if found_profile_id not in found_profiles:
                found_profiles[found_profile_id] = []

            # load up multiple options
            env_options = [profile['env_name'].lower(), profile['env_id'].lower(), profile['env_short_name'].lower()]

            if environment_name in env_options:
                found_profiles[found_profile_id].append(profile['env_id'])

        # let's first check to ensure we have only 1 profile
        if len(found_profiles) == 0:
            raise click.ClickException('no profile found with the provided application, environment, and profile names')
        if len(found_profiles) > 1:
            raise click.ClickException('multiple matching profiles found - cannot determine which profile to use')

        # and now we can check to ensure we have only 1 environment
        found_profile_id = list(found_profiles)[0]
        possible_environments = found_profiles[found_profile_id]
        if len(possible_environments) == 0:
            raise click.ClickException('no profile found with the provided application, environment, and profile names')
        if len(possible_environments) > 1:
            raise click.ClickException('multiple matching profiles found - cannot determine which profile to use')

        return {'profile_id': found_profile_id, 'environment_id': possible_environments[0]}

    @staticmethod
    def _validate_justification(justification: str):
        if justification and len(justification) > 255:
            raise ValueError('justification cannot be longer than 255 characters.')

    def ssh_aws_ssm_proxy(self, username, hostname, push_public_key, port_number, key_source):
        self.silent = True
        helper = hostname.split('.')
        instance_id = helper[0]
        aws_profile = None  # will drop to using standard aws boto3/cli profile provider chain
        aws_region = None  # will drop to using standard aws boto3/cli region provider chain
        if len(helper) > 1:
            aws_profile = helper[1]
        if len(helper) > 2:
            aws_region = helper[2]

        if push_public_key:
            details = self._ssh_generate_key(username=username, hostname=hostname, key_source=key_source)
            self._ssh_aws_push_key(
                instance_id=instance_id,
                aws_profile=aws_profile,
                aws_region=aws_region,
                username=username,
                key_pair=details['key_pair'],
            )

        commands = [
            'aws',
            'ssm',
            'start-session',
            f'--parameters portNumber={port_number}',
            '--document-name AWS-StartSSHSession',
            f'--target {instance_id}',
        ]

        if aws_profile:
            commands.append(f'--profile {aws_profile}')
        if aws_region:
            commands.append(f'--region {aws_region}')

        self.print(' '.join(commands), ignore_silent=True)

    @staticmethod
    def _ssh_generate_key_pair():
        # doing imports here as these packages are not a requirement to use pybritive in general
        from cryptography.hazmat.primitives.asymmetric import rsa
        from cryptography.hazmat.primitives import serialization

        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

        pem_private_key = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )

        pem_public_key = private_key.public_key().public_bytes(
            encoding=serialization.Encoding.OpenSSH, format=serialization.PublicFormat.OpenSSH
        )

        return {'private': pem_private_key, 'public': pem_public_key}

    def _ssh_generate_key(self, username, hostname, key_source):
        # doing imports here as these packages are not a requirement to use pybritive in general

        # these 3 ship with python3.x
        import glob
        import time
        import subprocess

        key_pair = self._ssh_generate_key_pair()

        # let's do the right thing and clean up old ephemeral keys
        ssh_dir = Path(self.config.path).parent.absolute() / 'ssh'
        ssh_dir.mkdir(exist_ok=True, parents=True)  # create the directory if it doesn't exist already
        if key_source == 'ssh-agent':
            # cleanup any old ssh keys that were randomly generated
            now = int(time.time())
            for key in glob.glob(f'{str(ssh_dir)}/random-*'):
                file = key.split('/')[-1].split('.')[0]
                expiration = int(file.split('-')[2])
                if expiration < now:
                    # Path(key).unlink(missing_ok=True)
                    # removed for now, for 3.7 compatability
                    try:
                        Path(key).unlink()
                    except FileNotFoundError:
                        pass

            pem_file = ssh_dir / f'random-{uuid.uuid4().hex}-{now + 60}.pem'
        elif key_source == 'static':
            # clean up the specific key if it exists, so we can create a new one
            pem_file = ssh_dir / f'{hostname}.{username}.pem'
            # pem_file.unlink(missing_ok=True)
            # removed for now, for 3.7 compatability
            try:
                pem_file.unlink()
            except FileNotFoundError:
                pass
        else:
            raise ValueError(f'invalid --key-source value {key_source}')

        # we only need to persist the private key locally
        # as the public key is just pushed to the ec2 instance
        # as a string in the ec2 instance connect api call (no file
        # reference)
        with open(str(pem_file), 'w', encoding='utf-8') as f:
            f.write(key_pair['private'].decode())
        os.chmod(pem_file, 0o400)

        # and if we are using ssh-agent we need to add the private key via ssh-add
        if key_source == 'ssh-agent':
            subprocess.run(['ssh-add', '-t', '60', '-q', str(pem_file)], check=False)

        return {'private_key_filename': pem_file, 'key_pair': key_pair}

    @staticmethod
    def build_import_exception_message(extras: str):
        return f'required packages not found. run `pip3 install pybritive[{extras}]`'

    @staticmethod
    def _ssh_aws_push_key(aws_profile, aws_region, instance_id, username, key_pair):
        try:
            import boto3
        except ImportError as e:
            raise click.ClickException(BritiveCli.build_import_exception_message('aws')) from e

        # we know we will be pushing the key to the instance so establish the
        # boto3 clients which are required to perform those actions
        session = boto3.Session(profile_name=aws_profile, region_name=aws_region)
        eic = session.client('ec2-instance-connect')
        ec2 = session.client('ec2')

        # now push the key
        az = ec2.describe_instances(InstanceIds=[instance_id])['Reservations'][0]['Instances'][0]['Placement'][
            'AvailabilityZone'
        ]

        eic.send_ssh_public_key(
            InstanceId=instance_id,
            InstanceOSUser=username,
            SSHPublicKey=key_pair['public'].decode(),
            AvailabilityZone=az,
        )

    def ssh_aws_openssh_config(self, push_public_key, key_source):
        lines = ['Match host i-*,mi-*']
        if push_public_key:
            commands = [
                '\tProxyCommand eval $(pybritive ssh aws ssm-proxy --hostname %h',
                '--username %r --port-number %p --push-public-key',
                f'--key-source {key_source})',
            ]
            lines.append(' '.join(commands))

            if key_source == 'static':
                ssh_dir = Path(self.config.path).parent.absolute() / 'ssh'
                lines.append(f'\tIdentityFile {str(ssh_dir)}/%h.%r.pem')
        else:
            line = '\tProxyCommand eval $(pybritive ssh aws ssm-proxy --hostname %h --username %r --port-number %p)'
            lines.append(line)

        self.print('Add the below Match directive to your SSH config file, after all Host directives.')
        self.print('This file is generally located at ~/.ssh/config.')
        self.print('Additional SSH config parameters can be added as required.')
        self.print('The below directive is the minimum required configuration.')
        self.print('')
        self.print('')
        self.print('\n'.join(lines))

    @staticmethod
    def aws_console(profile, duration, browser):
        # doing imports here as these packages are not a requirement to use pybritive in general

        # requests is a hard requirement for pybritive (via britive sdk)
        # and webbrowser ships with python3.x
        import requests
        import webbrowser

        # this is the one that may not be available so be careful
        try:
            import boto3
        except ImportError as e:
            raise click.ClickException(BritiveCli.build_import_exception_message('aws')) from e

        creds = boto3.Session(profile_name=profile).get_credentials()
        session_id = creds.access_key
        session_key = creds.secret_key
        session_token = creds.token
        json_creds = json.dumps({'sessionId': session_id, 'sessionKey': session_key, 'sessionToken': session_token})

        # Make request to AWS federation endpoint to get sign-in token. Construct the parameter string with
        # the sign-in action request and the JSON document with temporary credentials as parameters.

        params = {'Action': 'getSigninToken', 'SessionDuration': duration, 'Session': json_creds}

        url = 'https://signin.aws.amazon.com/federation'

        response = requests.get(url, params=params, timeout=15)

        signin_token = None
        try:
            signin_token = json.loads(response.text)
        except json.decoder.JSONDecodeError:
            raise click.ClickException(
                'Credentials have expired or another issue occurred. Please re-authenticate and try again.'
            )

        params = {
            'Action': 'login',
            'Issuer': 'pybritive',
            'Destination': 'https://console.aws.amazon.com/',
            'SigninToken': signin_token['SigninToken'],
        }

        # using requests.prepare() here to  help construct the url (with url encoding, etc.)
        # vs. doing it "manually" - we do not want to actually make a request to the url in python
        # but use the url to pop open a browser
        console_url = requests.Request('GET', url, params=params).prepare().url

        webbrowser.get(using=browser).open(console_url)

    def request_disposition(self, request_id, decision):
        self.login()

        if decision == 'approve':
            self.b.my_access.approve_request(request_id=request_id)
        if decision == 'reject':
            self.b.my_access.reject_request(request_id=request_id)

    def clear_cached_aws_credentials(self, profile):
        # start with the profile name that was passed in from the command
        Cache().clear_credentials(profile_name=profile)

        # then we can try to split it into parts and clear that version of the
        # profile name as well - it will not hurt anything to try to clear
        # both versions
        parts = self._split_profile_into_parts(profile)
        Cache().clear_credentials(profile_name=f"{parts['app']}/{parts['env']}/{parts['profile']}")

    def ssh_gcp_identity_aware_proxy(self, username, hostname, push_public_key, port_number, key_source):
        self.silent = True
        helper = hostname.split('.')
        instance_name = helper[1]
        project = helper[2]

        import subprocess
        import shlex

        command = f'gcloud compute instances list --format json --project {project}'
        instances = json.loads(subprocess.check_output(shlex.split(command)).decode('utf-8'))

        zone = None
        metadata = None
        for instance in instances:
            if instance['name'] == instance_name:
                zone = instance['zone'].split('/')[-1]
                metadata = instance.get('metadata')

        if not zone:
            raise click.BadParameter(f'no zone found for instance {instance_name} in project {project}')

        if push_public_key:
            details = self._ssh_generate_key(username=username, hostname=hostname, key_source=key_source)

            if push_public_key == 'os-login':  # this is pretty straightforward
                public_key = details['key_pair']['public'].decode('utf-8')
                command = (
                    f'gcloud compute os-login ssh-keys add --key="{public_key}" '
                    f'--project={project} --ttl=1m --no-user-output-enabled --quiet'
                )
                subprocess.check_output(shlex.split(command))
            if push_public_key == 'instance-metadata':  # this is a bit more involved
                now = datetime.now(timezone.utc).replace(microsecond=0)

                if metadata:
                    existing_keys_str = None
                    for item in metadata.get('items', []):
                        if item['key'] == 'ssh-keys':
                            existing_keys_str = item['value']
                            break
                    future_keys = []
                    if existing_keys_str:
                        existing_keys = existing_keys_str.split('\n')

                        for key in existing_keys:
                            should_carry_forward = True

                            if 'google-ssh' in key:  # non google-ssh keys should always carry forward
                                key_details = json.loads(key.split('google-ssh')[1].strip())
                                expires = key_details['expireOn']
                                if now > datetime.fromisoformat(expires):
                                    should_carry_forward = False
                            if should_carry_forward:
                                future_keys.append(key)

                    command = 'gcloud auth list --filter=status:ACTIVE --format="value(account)" --quiet'
                    active_gcloud_user = subprocess.check_output(shlex.split(command)).decode('utf-8').strip()

                    # format is 2023-05-16T20:15:22+0000
                    google_ssh_data = {
                        'userName': active_gcloud_user,
                        'expireOn': (now + timedelta(seconds=60)).strftime('%Y-%m-%dT%H:%M:%S+0000'),
                    }
                    google_ssh_data = json.dumps(google_ssh_data, separators=(',', ':'))  # separators IMPORTANT
                    public_key = details['key_pair']['public'].decode('utf-8')
                    new_key = f'{username}:{public_key} google-ssh {google_ssh_data}'
                    future_keys.append(new_key)

                    ssh_dir = Path(self.config.path).parent.absolute() / 'ssh'
                    ssh_dir.mkdir(exist_ok=True, parents=True)  # create the directory if it doesn't exist already
                    key_file = ssh_dir / uuid.uuid4().hex
                    with open(str(key_file), 'w', encoding='utf-8') as f:
                        f.write('\n'.join(future_keys))

                    commands = [
                        'gcloud',
                        'compute',
                        'instances',
                        'add-metadata',
                        instance_name,
                        '--project',
                        project,
                        '--metadata-from-file',
                        f'ssh-keys={str(key_file)}',
                        '--zone',
                        zone,
                        '--verbosity=error',
                        '--no-user-output-enabled',
                        '--quiet',
                    ]
                    subprocess.run(commands, check=False)
                    # key_file.unlink(missing_ok=True)
                    # removed for now, for 3.7 compatability
                    try:
                        key_file.unlink()
                    except FileNotFoundError:
                        pass

        commands = [
            'gcloud',
            'compute',
            'start-iap-tunnel',
            instance_name,
            port_number,
            '--listen-on-stdin',
            f'--project={project}',
            f'--zone={zone}',
            '--verbosity=error',
        ]

        self.print(' '.join(commands), ignore_silent=True)

    def ssh_gcp_openssh_config(self, push_public_key, key_source):
        lines = ['Match host gcp.*']
        if push_public_key:
            commands = [
                '\tProxyCommand eval $(pybritive gcp identity-aware-proxy --hostname %h',
                f'--username %r --port-number %p --push-public-key {push_public_key}',
                f'--key-source {key_source})',
            ]
            lines.append(' '.join(commands))

            if key_source == 'static':
                ssh_dir = Path(self.config.path).parent.absolute() / 'ssh'
                lines.append(f'\tIdentityFile {str(ssh_dir)}/%h.%r.pem')
        else:
            line = (
                '\tProxyCommand eval $(pybritive ssh gcp identity-aware-proxy --hostname %h --username %r '
                '--port-number %p)'
            )
            lines.append(line)

        self.print('Add the below Match directive to your SSH config file, after all Host directives.')
        self.print('This file is generally located at ~/.ssh/config.')
        self.print('Additional SSH config parameters can be added as required.')
        self.print('The below directive is the minimum required configuration.')
        self.print('')
        self.print('')
        self.print('\n'.join(lines))



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options
from ..helpers.profile_argument_decorator import click_smart_profile_argument


@click.command()
@build_britive
@britive_options(names='console,profile_type,tenant,token,silent,passphrase,federation_provider')
@click_smart_profile_argument
def checkin(ctx, console, profile_type, tenant, token, silent, passphrase, federation_provider, profile):
    """Checkin a profile.

    This command takes 1 required argument `PROFILE`. This should be a string representation of the profile
    that should be checked in. Format is `application name/environment name/profile name`.
    """
    ctx.obj.britive.checkin(
        profile=profile,
        console=console,
        profile_type=profile_type
    )



import click


option = click.option(
    '--console', '-c',
    default=False,
    is_flag=True,
    show_default=True,
    help='Checkout/checkin the console access for the profile instead of programmatic access.'
)



import click
from ..choices.output_format import output_format_choices


option = click.option(
    '--format', '-f', 'output_format',  # format is a reserved word so method parameter will be output_format
    default=None,
    help=(
        'Display output format. Valid values are (json, yaml, csv, table[-format]). '
        'If `table` is used an optional table format can be specified as `table-format`. '
        'Valid table formats can be found here: https://github.com/astanin/python-tabulate#table_format. '
        'Example: `table-pretty`.'
    ),
    show_choices=True,
    type=output_format_choices,
    show_default=True

)



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options


@click.group()
def ssh():
    """Facilitates connecting to private cloud servers using the standard ssh protocol."""
    pass


@ssh.group()
def aws():
    """Facilitates connecting, via SSH, to private AWS EC2 instances via Session Manager and EC2 Instance Connect.

    This command generally should be called from within an SSH config as opposed to being directly invoked.
    """
    pass


@ssh.group()
def gcp():
    """Facilitates connecting, via SSH, to private GCP Compute Instances via Identity Aware Proxy and OS Login/SSH Metadata.

    This command generally should be called from within an SSH config as opposed to being directly invoked.
    """
    pass


@aws.command()
@build_britive
@britive_options(names='ssh_push_public_key,ssh_key_source')
def config(ctx, push_public_key, key_source):
    """Prints the required Match directive to add to an OpenSSH config file (normally located at ~/.ssh/config.)"""

    ctx.obj.britive.ssh_aws_openssh_config(
        push_public_key=push_public_key,
        key_source=key_source
    )


@aws.command()
@build_britive
@britive_options(names='ssh_username,ssh_hostname,ssh_push_public_key,ssh_port,ssh_key_source')
def ssm_proxy(ctx, username, hostname, push_public_key, port_number, key_source):
    """Outputs AWS CLI command to be consumed by ProxyCommand to establish an OpenSSH proxy tunnel."""

    if push_public_key == 'default':
        push_public_key = True

    ctx.obj.britive.ssh_aws_ssm_proxy(
        username=username,
        hostname=hostname,
        push_public_key=push_public_key,
        port_number=port_number,
        key_source=key_source
    )


@gcp.command()
@build_britive
@britive_options(names='ssh_push_public_key,ssh_key_source')
def config(ctx, push_public_key, key_source):
    """Prints the required Match directive to add to an OpenSSH config file (normally located at ~/.ssh/config.)"""

    ctx.obj.britive.ssh_gcp_openssh_config(
        push_public_key=push_public_key,
        key_source=key_source
    )


@gcp.command()
@build_britive
@britive_options(names='ssh_username,ssh_hostname,ssh_push_public_key,ssh_port,ssh_key_source')
def identity_aware_proxy(ctx, username, hostname, push_public_key, port_number, key_source):
    """Outputs gcloud CLI command to be consumed by ProxyCommand to establish an OpenSSH proxy tunnel."""

    if push_public_key == 'default':
        push_public_key = 'os-login'

    ctx.obj.britive.ssh_gcp_identity_aware_proxy(
        username=username,
        hostname=hostname,
        push_public_key=push_public_key,
        port_number=port_number,
        key_source=key_source
    )



import base64
import configparser
import hashlib
import json
import os
from pathlib import Path
import random
import time
import webbrowser
import requests

from britive.britive import Britive
import click
from dateutil import parser
import jwt
from requests.adapters import HTTPAdapter, Retry
from .encryption import StringEncryption, InvalidPassphraseException


interactive_login_fields_to_pop = [
    'challengeParameters',
    'success',
    'type',
    'user',
    'userId',
    'username',
    'refreshToken',
    'authTime',
    'maxSessionTimeout'
]


federation_provider_default_expiration_seconds = 900


class InteractiveLoginTimeout(Exception):
    pass


def b64_encode_url_safe(value: bytes):
    return base64.urlsafe_b64encode(value).decode('utf-8').replace('=', '')


class CouldNotExtractExpirationTimeFromJwtException(Exception):
    pass


# this base class expects self.credentials to be a dict - so sub classes need to convert to dict
class CredentialManager:
    def __init__(self, tenant_name: str, tenant_alias: str, cli: any, federation_provider: str = None,
                 browser: str = os.getenv('PYBRITIVE_BROWSER')):
        self.cli = cli
        self.tenant = tenant_name
        self.alias = tenant_alias
        self.base_url = f'https://{Britive.parse_tenant(tenant_name)}'
        self.federation_provider = federation_provider
        self.browser = browser
        self.session = None

        # not sure if we really need 32 random bytes or if any random string would work
        # but the current britive-cli in node.js does it this way so it will be done the same
        # way in python
        while True:  # will break eventually when we get values that do not include --
            self.verifier = b64_encode_url_safe(bytes([random.getrandbits(8) for _ in range(0, 32)]))
            self.auth_token = b64_encode_url_safe(bytes(hashlib.sha512(self.verifier.encode('utf-8')).digest()))

            # WAF doesn't like to see `--` as it thinks it is a sql injection attempt
            if '--' not in self.verifier and '--' not in self.auth_token:
                break

        self.credentials = self.load() or {}

    def _setup_requests_session(self):
        self.session = requests.Session()
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[429, 500, 502, 503, 504])
        self.session.mount('https://', HTTPAdapter(max_retries=retries))
        global_ca_bundle = self.cli.config.get_tenant().get('ca_bundle')
        if global_ca_bundle:
            os.environ['PYBRITIVE_CA_BUNDLE'] = global_ca_bundle
            self.session.verify = global_ca_bundle
        # allow the disabling of TLS/SSL verification for testing in development (mostly local development)
        if os.getenv('BRITIVE_NO_VERIFY_SSL') and '.dev.' in self.tenant:
            # turn off ssl verification
            self.session.verify = False
            # wipe these due to this bug: https://github.com/psf/requests/issues/3829
            os.environ['CURL_CA_BUNDLE'] = ""
            os.environ['REQUESTS_CA_BUNDLE'] = ""
            # disable the warning message
            import urllib3
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    def perform_interactive_login(self):
        self.cli.print(f'Performing interactive login against tenant {self.tenant}.')
        sso_idp = self.cli.config.get_tenant().get('sso_idp')
        if sso_idp:
            sso_idp = sso_idp.replace('saml', 'SAML')  # ui is expecting it in caps
            url = f'{self.base_url}/sso?idp={sso_idp}&token={self.auth_token}'
        else:
            url = f'{self.base_url}/login?token={self.auth_token}'

        # establish a requests session which will be used in retrieve_tokens()
        self._setup_requests_session()

        try:
            webbrowser.get(using=self.browser).open(url)
        except webbrowser.Error:
            self.cli.print(
                'No web browser found. Please manually navigate to the link below and authenticate.'
            )
            self.cli.print(url)

        time.sleep(3)
        num_tries = 1
        while True:
            if num_tries > 60:
                raise InteractiveLoginTimeout()
            response = self.retrieve_tokens()

            if response.status_code >= 400:
                time.sleep(2)
                num_tries += 1
            else:
                credentials = response.json()['authenticationResult']

                try:
                    # attempt to pull the expiration time from the jwt
                    expiration_time_ms = self._extract_exp_from_jwt(
                        token=credentials['accessToken'],
                        verify=False,
                        convert_to_ms=True
                    )
                    self.cli.debug(f'found expiration time {expiration_time_ms} from jwt')
                except CouldNotExtractExpirationTimeFromJwtException:
                    # calculate from other fields in the authentication result
                    self.cli.debug('could not extract token expiration time from jwt - dropping to use other fields')
                    auth_time = int(credentials.get('authTime', 0))
                    session_time = int(credentials.get('maxSessionTimeout', 0))
                    expiration_time_ms = auth_time + session_time
                    self.cli.debug(f'found expiration time {expiration_time_ms} from authTime + maxSessionTimeout')

                credentials['safeExpirationTime'] = expiration_time_ms

                # drop a bunch of unnecessary fields
                for field in interactive_login_fields_to_pop:
                    credentials.pop(field, None)

                self.save(credentials)
                self.cli.print(f'Authenticated to tenant {self.tenant} via interactive login.')
                break

    @staticmethod
    def extract_field_from_jwt(token: str, field: str, verify: bool = False):
        try:
            return jwt.decode(
                token,
                # validation of the token will occur on the Britive backend
                # so not verifying everything here is okay since we are just
                # trying to extract the token expiration time so we can store
                # it in the ~/.britive/pybritive.credentials[.encrypted] file
                options={
                    'verify_signature': verify,
                    'verify_aud': verify
                }
            )[field]
        except Exception:
            return None

    @staticmethod
    def _extract_exp_from_jwt(token: str, verify: bool = False, convert_to_ms: bool = False):
        try:
            expiration_time = jwt.decode(
                token,
                # validation of the token will occur on the Britive backend
                # so not verifying everything here is okay since we are just
                # trying to extract the token expiration time so we can store
                # it in the ~/.britive/pybritive.credentials[.encrypted] file
                options={
                    'verify_signature': verify,
                    'verify_aud': verify
                }
            )['exp']
            return expiration_time * (1000 if convert_to_ms else 1)
        except Exception:
            raise CouldNotExtractExpirationTimeFromJwtException

    def perform_federation_provider_authentication(self):
        self.cli.print(f'Performing {self.federation_provider} federation provider authentication '
                       f'against tenant {self.tenant}.')

        # we need to extract the duration, if provided
        # field format is provider-[something provider specific]_[duration in seconds]
        helper = self.federation_provider.split('_')
        duration = federation_provider_default_expiration_seconds
        if len(helper) == 2:
            try:
                duration = int(helper[1])
            except ValueError:
                self.cli.print(f'Invalid federation provider duration {helper[1]} provided - defaulting '
                               f'to {duration} seconds.')

        # generate the token
        generated_token = Britive.source_federation_token_from(
            provider=helper[0],
            tenant=self.tenant,
            duration_seconds=duration
        )

        # obtain the provider and expiration time of the token
        provider, token = generated_token.split('::')
        provider = provider.lower()
        token = str(token)

        expiration_time = (int(time.time()) + federation_provider_default_expiration_seconds) * 1000

        try:
            if provider == 'aws':
                token = base64.b64decode(token.encode('utf-8'))
                token_expires = json.loads(token)['iam_request_headers']['x-britive-expires']
                expiration_time = int(parser.parse(token_expires).timestamp() * 1000)
            if provider == 'oidc':
                expiration_time = self._extract_exp_from_jwt(
                    token=token,
                    verify=False,
                    convert_to_ms=True
                )
        except Exception:
            self.cli.print(f'Cannot obtain token expiration time for {self.federation_provider}. Defaulting to '
                           f'{federation_provider_default_expiration_seconds} seconds.')

        # generate the credentials object and save it
        credentials = {
            'accessToken': generated_token,
            'safeExpirationTime': expiration_time
        }

        self.save(credentials)
        self.cli.print(f'Authenticated to tenant {self.tenant} via federation provider authentication.')

    def retrieve_tokens(self):
        url = f'{self.base_url}/api/auth/cli/retrieve-tokens'
        auth_params = {
            'authParameters': {
                'cliToken': self.verifier
            }
        }
        headers = {
            'Content-Type': 'application/json'
        }
        return self.session.post(url, headers=headers, json=auth_params)

    def load(self, full=False):
        # we should NEVER get here but adding here just in case
        raise click.ClickException('Must use a subclass of CredentialManager')

    def save(self, credentials: dict):
        # we should NEVER get here but adding here just in case
        raise click.ClickException('Must use a subclass of CredentialManager')

    def delete(self):
        # we should NEVER get here but adding here just in case
        raise click.ClickException('Must use a subclass of CredentialManager')

    # this helper exists since subclasses may need to override the method due to how the
    # access token may be encrypted and/or stored
    def _get_token(self):
        return self.credentials['accessToken']

    def get_token(self):
        if not self.has_valid_credentials():  # no credentials or expired creds for the tenant so do interactive login
            self.cli.debug('has_valid_credentials = False')
            # both methods below write the credentials out and update self.credentials as needed
            if self.federation_provider:
                self.perform_federation_provider_authentication()
            else:
                self.perform_interactive_login()

        token = self._get_token()
        return token

    def has_valid_credentials(self):
        if not self.credentials or self.credentials == {}:
            self.cli.print(f'Credentials for tenant {self.tenant} not found.')
            return False
        if int(time.time() * 1000) <= int(self.credentials.get('safeExpirationTime', 0)):
            self.cli.debug('credentials.py::has_valid_credentials - credentials exist and are not expired so are valid')
            return True
        self.cli.print(f'Credentials for tenant {self.tenant} have expired.')
        return False


class FileCredentialManager(CredentialManager):
    def __init__(self, tenant_name: str, tenant_alias: str, cli: any, federation_provider: str = None,
                 browser: str = os.getenv('PYBRITIVE_BROWSER')):
        home = os.getenv('PYBRITIVE_HOME_DIR', str(Path.home()))
        self.path = str(Path(home) / '.britive' / 'pybritive.credentials')
        super().__init__(tenant_name, tenant_alias, cli, federation_provider, browser)

    def load(self, full=False):
        path = Path(self.path)
        if not path.is_file():  # credentials file does not yet exist, create it as an empty file
            path.parent.mkdir(exist_ok=True, parents=True)
            path.write_text('', encoding='utf-8')

        # open the file with configparser
        credentials = configparser.ConfigParser()
        credentials.optionxform = str  # maintain key case
        credentials.read(str(path))
        credentials = json.loads(json.dumps(credentials._sections))  # TODO this is messy but works for now
        if full:
            return credentials
        return credentials.get(self.alias, None)

    def save(self, credentials: dict):
        full_credentials = self.load(full=True)
        if credentials is None:
            full_credentials.pop(self.alias, None)
            self.credentials = None
        else:
            full_credentials[self.alias] = credentials
            # effectively a deep copy
            self.credentials = json.loads(json.dumps(credentials))

        config = configparser.ConfigParser()
        config.optionxform = str  # maintain key case
        config.read_dict(full_credentials)

        # write the new credentials file
        with open(str(self.path), 'w', encoding='utf-8') as f:
            config.write(f, space_around_delimiters=False)

        jti = self.extract_field_from_jwt(
            token=(self.credentials or {}).get('accessToken'),
            verify=False,
            field='jti'
        )
        self.cli.debug(f'credentials.py::FileCredentialManager::save - set credentials to jwt id {jti}')

    def delete(self):
        self.save(None)


class EncryptedFileCredentialManager(CredentialManager):
    def __init__(self, tenant_name: str, tenant_alias: str, cli: any, passphrase: str = None,
                 federation_provider: str = None, browser: str = os.getenv('PYBRITIVE_BROWSER')):
        home = os.getenv('PYBRITIVE_HOME_DIR', str(Path.home()))
        self.path = str(Path(home) / '.britive' / 'pybritive.credentials.encrypted')
        self.passphrase = passphrase
        self.string_encryptor = StringEncryption(passphrase=self.passphrase)
        super().__init__(tenant_name, tenant_alias, cli, federation_provider, browser)

    def decrypt(self, encrypted_access_token: str):
        try:
            return self.string_encryptor.decrypt(ciphertext=encrypted_access_token)
        except InvalidPassphraseException:
            self.cli.print('invalid passphrase provided - wiping credentials and forcing a re-authentication.')
            self.delete()
            self.credentials = self.load() or {}
            return self.get_token()

    def encrypt(self, decrypted_access_token: str):
        return self.string_encryptor.encrypt(plaintext=decrypted_access_token)

    def load(self, full=False):
        path = Path(self.path)
        if not path.is_file():  # credentials file does not yet exist, create it as an empty file
            path.parent.mkdir(exist_ok=True, parents=True)
            path.write_text('', encoding='utf-8')

        # open the file with configparser
        credentials = configparser.ConfigParser()
        credentials.optionxform = str  # maintain key case
        credentials.read(self.path)
        credentials = json.loads(json.dumps(credentials._sections))  # TODO this is messy but works for now

        if full:
            return credentials
        return credentials.get(self.alias, None)

    def _get_token(self):
        # we should do a just-in-time decryption so unencrypted creds are not stored in memory
        return self.decrypt(self.credentials['accessToken'])

    def save(self, credentials: dict):
        full_credentials = self.load(full=True)
        if credentials is None:
            full_credentials.pop(self.alias, None)
            self.credentials = None
        else:
            credentials['accessToken'] = self.encrypt(credentials['accessToken'])
            full_credentials[self.alias] = credentials
            # effectively a deep copy
            self.credentials = json.loads(json.dumps(credentials))

        config = configparser.ConfigParser()
        config.optionxform = str  # maintain key case
        config.read_dict(full_credentials)

        # write the new credentials file
        with open(str(self.path), 'w', encoding='utf-8') as f:
            config.write(f, space_around_delimiters=False)

        jti = self.extract_field_from_jwt(
            token=(self.credentials or {}).get('accessToken'),
            verify=False,
            field='jti'
        )
        self.cli.debug(f'credentials.py::FileCredentialManager::save - set credentials to jwt id {jti}')

    def delete(self):
        self.save(None)



import click
from ..choices.browser import browser_choices


option = click.option(
    '--browser',
    type=browser_choices,
    default=None,
    show_choices=True,
    envvar='PYBRITIVE_BROWSER',
    show_envvar=True,
    help='The browser to use when opening a URL from the PyBritive CLI. Defaults to None which indicates the standard '
         'webbrowser selection process should be used. Can also source from PYBRITIVE_BROWSER.'
)



import click

option = click.option(
    '--duration', '-d',
    default=12 * 60 * 60,  # 12 hours in seconds (43200)
    show_default=True,
    help='Number of seconds the console session should last. Minimum of 900 and maximum of 43200 (the default).'
)



import click
import pkg_resources
from ..completers.profile import profile_completer
import os
import json


click_major_version = int(pkg_resources.get_distribution('click').version.split('.')[0])


def validate_profile(ctx, param, value):
    if 'KUBERNETES_EXEC_INFO' in os.environ:
        try:
            return json.loads(os.getenv('KUBERNETES_EXEC_INFO'))['spec']['cluster']['config']['britive-profile']
        except:
            raise ValueError('unable to find britive profile via cluster exec-extension with name britive-profile')
    return value


def is_required():
    return 'KUBERNETES_EXEC_INFO' not in os.environ


def click_smart_profile_argument(func):
    required = is_required()
    kwargs = {
        'required': required
    }
    if not required:
        kwargs['callback'] = validate_profile
    if click_major_version >= 8:
        kwargs['shell_complete'] = profile_completer

    dec = click.argument(
        'profile',
        **kwargs
    )
    return dec(func)



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options


@click.group()
def ls():
    """List resources available for currently authenticated identity."""
    pass


@ls.command()
@build_britive
@britive_options(names='format,tenant,token,silent,passphrase,federation_provider')
def applications(ctx, output_format, tenant, token, silent, passphrase, federation_provider):
    """List applications for the currently authenticated identity."""
    ctx.obj.britive.list_applications()


@ls.command()
@build_britive
@britive_options(names='format,tenant,token,silent,passphrase,federation_provider')
def environments(ctx, output_format, tenant, token, silent, passphrase, federation_provider):
    """List environments for the currently authenticated identity."""
    ctx.obj.britive.list_environments()


@ls.command()
@build_britive
@britive_options(names='checked_out,profile_type,output_format,tenant,token,silent,passphrase,federation_provider')
def profiles(ctx, checked_out, profile_type, output_format, tenant, token, silent, passphrase, federation_provider):
    """List profiles for the currently authenticated identity."""
    ctx.obj.britive.list_profiles(checked_out=checked_out, profile_type=profile_type)


@ls.command()
@build_britive
@britive_options(names='format,tenant,token,silent,passphrase,federation_provider')
def secrets(ctx, output_format, tenant, token, silent, passphrase, federation_provider):
    """List secrets for the currently authenticated identity."""
    ctx.obj.britive.list_secrets()


@ls.command()
@build_britive
@britive_options(names='format,tenant,token,silent,passphrase,federation_provider')
def approvals(ctx, output_format, tenant, token, silent, passphrase, federation_provider):
    """List approvals for the currently authenticated identity."""
    ctx.obj.britive.list_approvals()


@ls.command()
@build_britive
@britive_options(names='format,tenant,token,silent,passphrase,federation_provider')
def resources(ctx, output_format, tenant, token, silent, passphrase, federation_provider):
    """List resources for the currently authenticated identity."""
    ctx.obj.britive.list_resources()



import click
from ..choices.backend import backend_choices

option = click.option(
    '--backend', '-b', 'configure_backend',
    default=None,
    type=backend_choices,
    show_choices=True,
    show_default=True,
    help='The backend used to store temporary access tokens to authenticate against the Britive tenant.'
)



from britive.britive import Britive
from click.shell_completion import CompletionItem
import inspect


def api_completer(ctx, param, incomplete):
    # create an instance of the Britive class, so we can inspect it
    # this doesn't need to actually connect to any tenant, and we couldn't even if we
    # wanted to since when performing shell completion we have no tenant/token
    # context in order to properly establish a connection.
    b = Britive(token='ignore', tenant='britive.com', query_features=False)

    # parse the incomplete command, so we can determine where in the "hierarchy" we are
    # and what commands/subcommands the user should be presented with
    parts = incomplete.split('.')[:-1]
    not_base_level = len(parts) > 0
    for part in parts:
        b = getattr(b, part)

    existing = '.'.join(parts)

    options = []

    # vars happen at all levels
    for var, value in vars(b).items():
        # filter out things which should not show as completion items
        if not str(value).startswith('<britive.') or var == 'britive':
            continue

        method = f'{existing}.{var}' if not_base_level else var

        if method.lower().startswith(incomplete.lower()):
            doc_line = f'methods related to {var}'
            try:
                doc_line = inspect.getdoc(getattr(b, var)).split('\n')[0]
            except:
                pass
            options.append(CompletionItem(method, help=doc_line))

    # dir only happens at non "base" levels
    if not_base_level:
        # for each method in the class
        for func in dir(b):
            # filter out methods which are not callable and methods which are not "public"
            if not callable(getattr(b, func)) or func.startswith("_"):
                continue

            method = f'{existing}.{func}'

            # if this method is a potential match add it to the completion list
            if method.lower().startswith(incomplete.lower()):
                # grab the doc string if present
                doc_line = f'no docs found for {method}'

                try:
                    doc_line = inspect.getdoc(getattr(b, func)).split('\n')[0]
                except:
                    pass

                options.append(CompletionItem(method, help=doc_line))
    return options




from pathlib import Path


def common_asserts(result, substring=None, exit_code=0):
    assert result.exit_code == exit_code
    if isinstance(substring, str):
        substring = [substring]
    for sub in substring:
        assert sub in result.output


def test_view(runner, cli):
    result = runner.invoke(cli, 'secret view /pybritive-test-standard -f yaml'.split(' '))
    common_asserts(result, 'test')


def test_download(runner, cli):
    filename = 'pybritive-test-secret-file.txt'
    result = runner.invoke(cli, 'secret download /pybritive-test-file'.split(' '))
    message = 'wrote contents of secret file to'
    common_asserts(result, message)
    with open(filename, 'r', encoding='utf-8') as f:
        assert 'test' in f.read()
    path = Path(filename)
    # path.unlink(missing_ok=True)
    # removed for now, for 3.7 compatability
    try:
        path.unlink()
    except FileNotFoundError:
        pass


def test_download_filename_provided(runner, cli):
    filename = 'pybritive-test-secret-file-2.txt'
    result = runner.invoke(cli, f'secret download /pybritive-test-file -F {filename}'.split(' '))
    message = 'wrote contents of secret file to'
    common_asserts(result, message)
    with open(filename, 'r', encoding='utf-8') as f:
        assert 'test' in f.read()
    path = Path(filename)
    # path.unlink(missing_ok=True)
    # removed for now, for 3.7 compatability
    try:
        path.unlink()
    except FileNotFoundError:
        pass



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options


@click.command()
@build_britive
@britive_options(names='tenant,silent,passphrase')
def logout(ctx, tenant, silent, passphrase):
    """Logout of an interactive login session.

    This only applies when an API token has not been specified via `--token,-T` or via environment variable
    `BRITIVE_API_TOKEN`.
    """
    ctx.obj.britive.logout()



import click
from ..choices.ssh_key_source import ssh_key_source_choices


option = click.option(
    '--key-source', '-k',
    type=ssh_key_source_choices,
    show_choices=True,
    required=True,
    default='ssh-agent',
    show_default=True,
    help='The source of the SSH key used to authenticate to the remove server.'
)
